
check.conv.optimx <- function(obj){

  obj1 <- obj[(obj$kkt1 + obj$kkt2 == 2) & !is.na(obj$kkt1 + obj$kkt2), ]
	obj1[order(obj1$value), ]
}

maop <- function(formula,
                 subjects,
                 periods=subjects,
                 randomEffect=FALSE,
                 dynamics=NULL,
                 startValues=NULL,
                 distance=2,
                 print.info=TRUE,
                 RBR=FALSE, 
                 reparam=TRUE, 
                 approx_hessian=FALSE,
                 methods=c("Nelder-Mead", "BFGS"), all.methods=FALSE, grad.tol.conv=1e-2,...)
{
  thisCall <- match.call()
  ## check if the intercept is present
  if(!attr(terms(formula), "intercept"))
    stop("\nError: the model requires an intercept.\n")
  if(is.null(subjects))
    stop("\nError: group variable <subjects> not specified.\n")
  ## possible different dynamics 
  if(!is.null(dynamics)){
    if(!is.factor(dynamics))
      stop("\nError: dynamics must be a factor.\n")
    else
      numCorr <- nlevels(dynamics)
  }
  else
    numCorr <- 1
  ## check if valid periods
  if(!identical(length(subjects), length(periods)))
    stop("\nError: subjects and periods should have the same length.\n")
  if(nlevels(subjects)<2)
    stop("\nError: at least two subjects should be present.\n")
  if(nlevels(subjects)>nlevels(periods))
    stop("\nError: more subjects than measurament periods.\n")
  ## check valid distance
  if(distance<1)
    stop("\nError: distance should be positive and larger than one.\n")

  thisFrame <- model.frame(formula, na.action=na.pass)
  response <- model.response(thisFrame)
  ## check if the response is ordinal
  if(!is.ordered(response))
    stop("\nError: the response must be an ordinal categorical variable.\n")
  numCateg <- nlevels(as.factor(response))
  if(numCateg==1)
    stop("\nError: at least two categories must be present.\n")
  designMatrix <- model.matrix(formula, thisFrame)
  numRegr <- dim(designMatrix)[2]
  numPar <- (numCateg-2)+numRegr+numCorr+randomEffect

	ind.startValues <- 0
  if(!is.null(startValues))
    if(length(startValues)==numPar)
      startTheta <- startValues
    else
      stop("\nError: number of starting values not suitable.\n")
  else{
    ind.startValues <- 1
    ## compute starting values using a standard (non-dynamic)
    ## cumulative probit model by polr() in MASS package
    if(print.info)
      cat("\nComputing starting values...\n")
    if(numCateg>2){
      require(MASS)
      po <- polr(formula, method="probit", na.action=na.omit)
      if(length(po$coeff)!=(numRegr-1) | length(po$zeta)!=(numCateg-1))
        stop("\nError: polr() routine unables to find suitable starting values.\n")
      startTheta <- c(po$zeta[2:(numCateg-1)]-po$zeta[1], -po$zeta[1], po$coeff,
                       rep(0.0, numCorr))
    }
    else{ ## two categories
      po <- glm(formula, family=binomial(link="probit"), na.action=na.omit)
      startTheta <- c(po$coefficients, rep(0.0, numCorr))
    }
    if(randomEffect)
      startTheta <- c(startTheta, 0.1)
  }
  ## names
  names(startTheta)[(numCateg-1)] <- "intercept"
  if(numCorr==1)
    names(startTheta)[numCateg+numRegr-1] <- "autocorrelation"
  else
    for(i in 1:numCorr)
      names(startTheta)[numCateg+numRegr-2+i] <-
        paste(paste("autocorr", ":", sep=""), levels(dynamics)[i], sep="")
  if(randomEffect)
    names(startTheta)[length(startTheta)] <- "random effect"

  if(print.info){
 		if(!RBR)
	    cat("\nComputing maximum pairwise likelihood estimators...\n")
		else
	    cat("\nComputing maximum penalized pairwise likelihood estimators...\n")
	}

  ## some operations to prepare the data
  ## transform NA's values for compatibility with my (poor) C code!
  response <- as.numeric(response)
  response[is.na(response)] <- -999
  jumps <- diff(as.numeric(periods))
  whereJumps <- which(jumps!=0)
  lengths <- diff(whereJumps)
  dimData <- c(whereJumps[1], lengths, length(response)-sum(lengths)-whereJumps[1])
  ## create period-specific indicator for different dynamics
  if(!is.null(dynamics))
    dynCateg <- as.integer(c(dynamics[1], dynamics[whereJumps+1])) ## CORREZIONE +1 28/8/2008
  else 
    dynCateg <- rep(1, length(dimData))
  ## create period-specific indicator for different subjects
  idSubj <- as.integer(c(subjects[1], subjects[whereJumps+1])) ## CORREZIONE +1 28/8/2008
  idSubj.f <- as.factor(idSubj)
  numSubj <- nlevels(idSubj.f)
  ## new labels for subjects following increasing numeric order
  levels(idSubj.f) <- as.character(1:numSubj)
  idSubj <- as.numeric(idSubj.f)
  numPairs <- rep(0, numSubj)
  ## read dimensions and data in the C code
   .C("read_dimensions",
     as.integer(numCateg),
     as.integer(numRegr),
     as.integer(numCorr),
     as.integer(randomEffect),
     as.integer(dimData),
     as.integer(sum(dimData)),
     as.integer(length(dimData)),
     as.integer(numSubj),
      package="maop")
  .C("read_order_pair_lik",
     as.integer(distance),
     package="maop")
  .C("allocate",
     package="maop")
  .C("read_data",
     as.integer(response),
     package="maop")
  .C("read_covariates",
     as.double(designMatrix),
     package="maop")
  .C("read_dynamic_categories",
     as.integer(dynCateg),
     package="maop")
   .C("read_index_subjects",
     as.integer(idSubj),
	   package="maop")
  .C("read_approx_hessian",
     as.integer(approx_hessian),
     package="maop")

  logit <- function(x, min=0, max=1){
    x <- (x - min)/(max - min)
    log(x/(1 - x))
  }
  inv.logit <- function(x, min=0, max=1){
    p <- exp(x)/(1 + exp(x))
    p <- ifelse(is.na(p) & !is.na(x), 1, p)
    p * (max - min) + min
  }
  transf <- function(x, eps=0){
    y <- x
    if(numCateg>2){
      cutpoints <- c(0, x[1:(numCateg-2)])
      y[1:(numCateg-2)] <- log(diff(cutpoints)-eps)
    }
    y[(numPar-numCorr):(numPar-randomEffect)] <-
      logit(x[(numPar-numCorr):(numPar-randomEffect)], min=-1+eps, max=1-eps)
    if(randomEffect)
      y[numPar] <- log(x[numPar]-eps)
    y
  }
  back.transf <- function(y, eps=0){
    x <- y
    if(numCateg>2){
      x[1:(numCateg-2)] <- cumsum(exp(y[1:(numCateg-2)])+eps)
    }
    x[(numPar-numCorr):(numPar-randomEffect)] <- inv.logit(y[(numPar-numCorr):(numPar-randomEffect)], min=-1+eps, max=1-eps)
    if(randomEffect)
      x[numPar] <- exp(y[numPar])+eps
    x
  }
  jacob <- function(y, eps=0){
    J <- diag(numPar)
    exp.y <- exp(y)
    if(numCateg>2){
      mat <- diag(numCateg-2)
      mat[lower.tri(mat, diag=TRUE)] <- rep(exp.y[1:(numCateg-2)], (numCateg-2):1)
      J[1:(numCateg-2),1:(numCateg-2)] <- mat
    }
    start <- numPar-numCorr
    end <- numPar-randomEffect
    for(i in start:end)
      J[i,i] <- 2*(1-eps)*exp.y[i]/(1+exp.y[i])^2
    if(randomEffect)
      J[numPar,numPar] <- exp.y[numPar]
    J
  }
  logPair <- function(y, transf=TRUE){
    if(transf) x <- back.transf(y)
    else x <- y
    if(any(!is.finite(x)))
      NA
    else{
      .C("logPair",
         as.double(x),
         logPair=as.double(0.0),
         package="maop")$logPair
    }
  }
  score.logPair <- function(y, transf=TRUE){
    if(transf) x <- back.transf(y)
    else x <- y
    if(any(!is.finite(x)))
      rep(NA, numPar)
    else{
      tmp <- .C("score_info_logPair",
                as.double(x),
                score=as.double(rep(0.0, numPar)),
                var.score=as.double(rep(0.0, numPar^2)),
                info=as.double(rep(0.0, numPar^2)),
                package="maop")
      if(transf)
        t(jacob(y))%*%tmp$score
      else
        tmp$score
    }
  }  
  
    scoreInfo.logPair <- function(y, approx_hessian, transf=TRUE){
    if(transf) x <- back.transf(y)
    else x <- y
    if(any(!is.finite(x))){
      tmp <- list()#list(score, var.score, info)
      tmp$score <- rep(NA, numPar)
      tmp$var.score <- matrix(NA, ncol=numPar, nrow=numPar)
      tmp$info <- matrix(NA, ncol=numPar, nrow=numPar)
    }else{
      tmp <- .C("score_info_logPair",
                as.double(x),
                score=as.double(rep(0.0, numPar)),
                var.score=as.double(rep(0.0, numPar^2)),
                info=as.double(rep(0.0, numPar^2)),
                approx_hessian=as.integer(approx_hessian),
                package="maop")
      tmp$info <- matrix(tmp$info, ncol=numPar, byrow=TRUE)
      tmp$var.score <- matrix(tmp$var.score, ncol=numPar, byrow=TRUE)
      if(transf){

				if(!approx_hessian){
				
					D2Jacob <- lapply(1:length(y), function(u) hessian(x=y, function(a) back.transf(a)[u]))
					for(i in 1:length(tmp$score))
						D2Jacob[[i]] <- tmp$score[i]*D2Jacob[[i]]

		      J <- jacob(y)
		      tmp$score <- t(J)%*%tmp$score
		      tmp$var.score <- t(J)%*%tmp$var.score%*%J

		      tmp$info <- t(J)%*%tmp$info%*%J + Reduce("+", D2Jacob)
				}else{

				  J <- jacob(y)
				  tmp$score <- t(J)%*%tmp$score
				  tmp$var.score <- t(J)%*%tmp$var.score%*%J
				  tmp$info <- t(J)%*%tmp$info%*%J
				
				}

      }
    }
 			tmp
  }

		logPair.RBR <- function(y, approx_hessian, transf=TRUE){#, print.hessians=FALSE){
			pllik <- logPair(y, transf)
		  tmp <- scoreInfo.logPair(y, approx_hessian, transf)
		  H <- -tmp$info
		  J <- tmp$var.score
			trace.adj <- sum(solve(H)*J)
#				if(print.hessians){
#					print("numerical hessian")
#					print(hessian(x=y, function(u) logPair(u, transf)))
#					print("analytical hessian")
#					print(-H)
#				}
				
			pllik - 0.5*trace.adj
		}


	if(!RBR){
		pair <- optimx(transf(startTheta), fn=function(u) -logPair(u), gr=function(u) -score.logPair(u), method=methods, control=list(all.methods=all.methods, ...))
		pair.whole.out <- pair
		#take the best
		pair <- check.conv.optimx(pair)
		if(nrow(pair)==0){
			pair <- list()
			pair$convergence <- 1
		}else{	
			pair <- pair[1,]			
			theta <- as.numeric(pair[1:numPar])
				if(!reparam)
					theta <- back.transf(theta)
				
			score_at_max <- attributes(pair)$details[1,]$ngatend#grad(x=theta, function(u) logPair(u, reparam)) 
			if(max(abs(score_at_max))>=grad.tol.conv)
				pair$convergence <- 1
			else
				pair$convergence <- 0
		}
	}else{

			##if starting values are supplied, then it is supposed that they are given in the right parametrization
			##if starting values are computed by maop function, then it we need to transform them whenever reparam=TRUE
 		  if(reparam & ind.startValues)
	 		  startTheta <- transf(startTheta)
			
#			pair <- optimx(startTheta, fn=function(u) -logPair.RBR(u, reparam, print.hessians), method=methods, control=list(all.methods=all.methods, ...))
			pair <- optimx(startTheta, fn=function(u) -logPair.RBR(u, reparam), method=methods, control=list(all.methods=all.methods, ...))
			pair.whole.out <- pair
			#take the best
			pair <- check.conv.optimx(pair)
			if(nrow(pair)==0){
				pair <- list()
				pair$convergence <- 1
			}else{	
				pair <- pair[1,]			
				theta <- as.numeric(pair[1:numPar])
				score_at_max <- score_at_max <- attributes(pair)$details[1,]$ngatend#grad(x=theta, function(u) logPair.RBR(u, reparam)) 
				if(max(abs(score_at_max))>=grad.tol.conv)
					pair$convergence <- 1
				else
					pair$convergence <- 0

			}
	}
 
 
  if(pair$convergence!=0){
    on.exit(.C("deallocate",
               package="maop"))

     theta <- seTheta <- rep(NA, numPar)
     varTheta <- matrix(NA, numPar, numPar)
     pair$value <- NA
     clic <- NA

    warning("\nWarning: not possible to compute the maximum pairwise likelihood estimator. The optimizer did not reach the convergence!\n")
  }else{
    if(print.info)
      cat("\nComputing standard errors...\n")

    tmp <- scoreInfo.logPair(theta, approx_hessian, reparam)
    H <- tmp$info
    H.inv <- solve(H)
    J <- (numSubj)/(numSubj-numPar)*tmp$var.score
    ## usual sandwich variance estimate 
    varTheta <- H.inv%*%J%*%H.inv
    seTheta <- sqrt(diag(varTheta))
    clic <- -2*pair$value+2*sum(diag(J%*%H.inv))
    on.exit(.C("deallocate",
               package="maop"))
    names(seTheta) <- names(startTheta)
    names(theta) <- names(startTheta)
  }
  
  structure(list(formula=formula,
                 call=thisCall,
                 subjects=subjects,
                 periods=periods,
                 randomEffect=randomEffect,
                 dynamics=dynamics,
                 distance=distance,
                 startValues=startTheta,
                 coeff=theta,
                 stdErr=seTheta,
                 varMatrix=varTheta,
                 logPair=-pair$value,
                 clic=-clic,
                 convergence=pair$convergence,
                 opt.pair=pair.whole.out, 
                 RBR=RBR), class="maop")
}


print.maop <- function(x, digits = NULL, ...)
{

 if(is.null(digits))
   digits <- options()$digits
 else
   options(digits = digits)
  coeff <- cbind(x$coeff, x$stdErr, abs(x$coeff)/x$stdErr)
  dimnames(coeff) <- list(names(x$coeff), c("Value", "Std. Error", "|z value|"))
  x$coeff <- coeff
  
  cat("\nMarginal autoregressive ordered probit model.\n")
  cat("\nCall:\n")
  print(x$formula)
  if(isTRUE(x$isRandomEffect))
	cat("\nPatient-specific random effect included\n")
  if(!is.null(x$dynanics))
    cat("\nAutocorrelation described by ", x$dynamics, "\n") 
	if(!x$RBR)
	 cat("\nMaximum pairwise likelihood estimates with maximal distance ", x$distance, ": \n")
	else
		 cat("\nMaximum penalized pairwise likelihood estimates with maximal distance ", x$distance, ": \n")

  print(x$coeff, digits=digits)
  if(x$convergence!=0)
    cat("\nWARNING: optimization routine `optimx` did not converge\n")
#    cat("WARNING: Optim converge code", x$convergence, "; see ?optim for details\n")
  cat("\nLog-pairwise likelihood:\n")
  print(x$logPair, digits=digits)
  cat("\nCLIC statistic:\n")
  print(x$clic, digits=digits)
  
 invisible(x)
}


summary.maop <- function(object, ...)
{
  print.maop(object, ...)
}
