#if 1
//version of 12/08/2019
//In this version of the code the function
//NL_score_info_log_prob_pair
//computes the exact hessian rather than an approximation


#include <R.h>
#include <Rmath.h>
#include <R_ext/RS.h>
#include <R_ext/Applic.h>
#include "pair_maop.h"
#define MISSING_VALUE -999  /* here, -999 means NA! */
#define IS_MISSING(data) ((data) == MISSING_VALUE ? 1 : 0) 
#define SMALL 1e-4
#define BIG 1e+8

#define NL 0
#if  NL == 1
	#define CV 0
#else
	#define CV 1
#endif

/********************************************************************************
STATIC VARIABLES
********************************************************************************/

/* variables containing data and design matrix */
static int *data;
static double **design_matrix;

/* variable containing the autocorrelation categories */
static int *dynamic_categ;

/* variable containing the period-specific subject indeces */
static int *index_subjects;

/* variable containing information wheter approximate hessian should be computed */
//added by NL on 12/31/2019
int approx_hessian;


/* structure containing the parameters */
struct Parameters
{
  double *tau;
  double *beta;
  double *phi;
  double sigma2;
};

static struct Parameters Param;

/* structure containing the data "dimensions" */
struct Dimensions
{
  int num_categ; /* number of categories */
  int num_regr;  /* number of regressors (intercept included) */
  int num_auto_corr; /* number of autocorrelation parameters */
  int is_random_effect; /* dummy for presence random effect */
  int *data;     /* number of observations for each measurement period */
  int tot_data;  /* total number of observations */ 
  int num_periods;  /* number of measurement periods */ 
  int num_subjects;  /* number of subjects */  
  int num_param;   /* number of parameters */
};

static struct Dimensions Dim;

/* variables used for evaluating the log-pairwise likelihood */
static int index_dynamic;
static int order_pair;
static int start_index;
static int end_index;
static int this_subject;

static double *means;
static double *logPair_subject;
static double **score_subject;
static double **info_subject;
static double *score_period;
static double *info_period;
static double *score_this_pair;
static double *info_this_pair;

void allocate()		   
{

  int i;
  int dim_info=Dim.num_param*Dim.num_param;
  
  /* allocate the data */
  data=Calloc(Dim.tot_data, int);

  /* allocate the design matrix */
  design_matrix=Calloc(Dim.tot_data, double*);
  for(i=0; i<Dim.tot_data; i++)
    design_matrix[i]=Calloc(Dim.num_regr, double);
  
  /* allocate dynamic categories */
  dynamic_categ=Calloc(Dim.num_periods, int);

  /* allocate subject indeces */
  index_subjects=Calloc(Dim.num_periods, int);

  /* dynamic allocation of the thresholds vector */
  Param.tau=Calloc((Dim.num_categ-1), double);
  /* dynamic allocation of regressor param's */
  Param.beta=Calloc(Dim.num_regr, double); 
  /* dynamic allocation of the autocorrelation param's */
  Param.phi=Calloc(Dim.num_auto_corr, double);

  /* dynamic allocation of auxiliar variables for 
     pairwise likelihood evaluation */

  means=Calloc(Dim.tot_data, double);
  
  logPair_subject=Calloc(Dim.num_subjects, double);
  score_subject=Calloc(Dim.num_subjects, double*);
  info_subject=Calloc(Dim.num_subjects, double*);
  for(i=0; i<Dim.num_subjects; i++){
    score_subject[i]=Calloc(Dim.num_param, double);
    info_subject[i]=Calloc(dim_info, double);
  }
  score_period=Calloc(Dim.num_param, double);
  score_this_pair=Calloc(Dim.num_param, double);
  info_period=Calloc(dim_info, double);
  info_this_pair=Calloc(dim_info, double);

} 

void deallocate()
{
  int i;
   
  Free(data);
  Free(dynamic_categ);
  Free(index_subjects);
  Free(Param.beta);
  Free(Param.tau);
  Free(Param.phi);
  Free(Dim.data);
  
  for(i=0; i<Dim.tot_data; i++)
    Free(design_matrix[i]);
  Free(design_matrix);

  Free(means);
  
  Free(score_this_pair);
  Free(score_period);
  Free(info_this_pair);
  Free(info_period);
  for(i=0; i<Dim.num_subjects; i++){
    Free(score_subject[i]);
    Free(info_subject[i]);
  }
  Free(score_subject);
  Free(info_subject);
  Free(logPair_subject);
}

void read_dimensions(const int *num_categ,
		     const int *num_regr,
		     const int *num_auto_corr,
		     const int *is_random_effect,
		     const int *dim_data, 
		     const int *dim_tot_data,
		     const int *num_periods,
		     const int *num_subjects)
{

  int i;

  Dim.num_param=(*num_categ-2)+*num_regr+*num_auto_corr+*is_random_effect;
  
  Dim.num_categ=*num_categ;
  Dim.num_regr=*num_regr;
  Dim.num_auto_corr=*num_auto_corr;
  Dim.is_random_effect=*is_random_effect;
  Dim.tot_data=*dim_tot_data;
  Dim.num_periods=*num_periods;
  Dim.num_subjects=*num_subjects;

  Dim.data=Calloc(Dim.num_periods, int);
  for(i=0; i<Dim.num_periods; i++)
    Dim.data[i]=dim_data[i];

}

void read_order_pair_lik(const int *order_pair_lik)
{

  order_pair=(*order_pair_lik);

}

void read_data(const int *data_input)
{

  int i;
  for(i=0; i<Dim.tot_data; i++)
    data[i]=data_input[i];
  
}

void read_covariates(const double *design_matrix_input)
{

  int i,j;

  for(i=0; i<Dim.tot_data; i++)
    for(j=0; j<Dim.num_regr; j++)
      design_matrix[i][j]=design_matrix_input[i+j*Dim.tot_data];
    
}

void read_dynamic_categories(const int *dynamic_categ_input)
{

  int i;
  
  for(i=0; i<Dim.num_periods; i++)
    /* for compatibility with C code categories must be 
       shifted back of one unit */
    dynamic_categ[i]=dynamic_categ_input[i]-1;

}

void read_index_subjects(const int *index_subjects_input)
{

  int i;
  
  for(i=0; i<Dim.num_periods; i++)
    /* for compatibility with C code indeces must be 
       shifted back of one unit */
    index_subjects[i]=index_subjects_input[i]-1;
}

/* this is called only inside C differently from the previous "read_" functions! */
static void read_parameters(const double *param)
{
  int i, j;

  Param.tau[0]=0.0;
  for(i=1; i<(Dim.num_categ-1); i++)
    Param.tau[i]=param[i-1];
  for(i=0; i<Dim.num_regr; i++)
    Param.beta[i]=param[i+Dim.num_categ-2];
  /* compute all the means */
  for(i=0; i<Dim.tot_data; i++){
    means[i]=0.0;
    for(j=0; j<Dim.num_regr; j++){
      means[i]+=design_matrix[i][j]*Param.beta[j];
    }
  }
  for(i=0; i<Dim.num_auto_corr; i++)
    Param.phi[i]=param[Dim.num_categ+Dim.num_regr-2+i];
  if(Dim.is_random_effect)
    Param.sigma2=param[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr];
  else /* any random effect */
    Param.sigma2=0.0;   
}


//by NL on 12/31/2019
void read_approx_hessian(const int *ah){
	approx_hessian = *ah;
}


static void build_limits(const int data,
			 const double regr, /* regression component! */
			 const double sd,
			 double* lim_inf,
			 double* lim_sup,
			 int* infin)
{
	
  if (data>1 && data<Dim.num_categ){
    *lim_inf=(Param.tau[data-2]-regr)/sd;
    *lim_sup=(Param.tau[data-1]-regr)/sd;
    *infin=2;
  }
  else if (data==1){
    *lim_inf=-BIG;  /* say, minus infinity */
    *lim_sup=(Param.tau[0]-regr)/sd;
    *infin=0;
  }
  else if (data==Dim.num_categ){
    *lim_inf=(Param.tau[Dim.num_categ-2]-regr)/sd;  
    *lim_sup=BIG;  /* say, plus infinity */
    *infin=1;
  }

}

void logPair(const double *param,
	     double *logPair)
{

  int i;
  double sum;

  read_parameters(param);

  for(i=0; i<Dim.num_subjects; i++)
    logPair_subject[i]=0.0;
  
  start_index=0;
  end_index=0;
  this_subject=0;
  for(i=0; i<Dim.num_periods; i++){
    start_index=end_index;
    end_index+=Dim.data[i];
    index_dynamic=dynamic_categ[i];
    this_subject=index_subjects[i];
    logPair_period();
  }
  sum=0.0;
  for(i=0; i<Dim.num_subjects; i++)
    sum+=logPair_subject[i]/Dim.num_subjects;
  *logPair=sum;
  
}

/* compute score and Fisher info */
void score_info_logPair(const double *param,
			double *tot_score,
			double *tot_var_score,
			double *tot_info,
			int approx_hessian)
{
  int i, j, k;

  read_parameters(param);

  /* 1st step compute subject-specific pieces */
  for(k=0; k<Dim.num_subjects; k++){
    for(i=0; i<Dim.num_param; i++)
      score_subject[k][i]=0.0;
    for(i=0; i<Dim.num_param; i++)
      for(j=0; j<Dim.num_param; j++)
	info_subject[k][i+j*Dim.num_param]=0.0;
  }
  this_subject=0;
  start_index=0;
  end_index=0;
  for(i=0; i<Dim.num_periods; i++){
    start_index=end_index;
    end_index+=Dim.data[i];
    index_dynamic=dynamic_categ[i];
    this_subject=index_subjects[i];
    score_info_logPair_period(param);
  }


	int n_subj_squared = R_pow_di(Dim.num_subjects, 2);
  /* 2nd step aggregate subject-specific pieces */
  for(i=0; i<Dim.num_param; i++)
    tot_score[i]=0.0;
  for(i=0; i<Dim.num_param; i++)
    for(j=0; j<Dim.num_param; j++)
      tot_info[i+j*Dim.num_param]=0.0;
  for(i=0; i<Dim.num_param; i++)
    for(j=0; j<Dim.num_param; j++)
      tot_var_score[i+j*Dim.num_param]=0.0;

  for(k=0; k<Dim.num_subjects; k++){
    /* update global vectors/matrices */
    for(i=0; i<Dim.num_param; i++)
      tot_score[i]+=score_subject[k][i]/Dim.num_subjects;
    for(i=0; i<Dim.num_param; i++)
      for(j=0; j<Dim.num_param; j++)
	tot_info[i+j*Dim.num_param]+=
	  info_subject[k][i+j*Dim.num_param]/Dim.num_subjects;
      for(i=0; i<Dim.num_param; i++)
	for(j=0; j<Dim.num_param; j++)
	  tot_var_score[i+j*Dim.num_param]+=
	    score_subject[k][i]*score_subject[k][j]/n_subj_squared;
  }

}



static void logPair_period()
{

  int t, j;
  double result=0.0;
    
  for (t=(start_index+1); t<end_index; t++){
    /* if the data at time t is not missing... */
    if(!IS_MISSING(data[t])){
      for(j=1; j<=imin2(order_pair, t-start_index); j++){
	if(!IS_MISSING(data[t-j])){
	  prob_pair(t, j, &result);
	  logPair_subject[this_subject]+=log(result);
	}
      }
    }
  }
}


static void prob_pair(const int position,
		      const int distance_neigh,
			   double *result)
{


  int t=position;
  int l=distance_neigh;
  double this_phi=Param.phi[index_dynamic];

  double var=Param.sigma2+1.0;
  double sd=sqrt(var);
  double cov=Param.sigma2+R_pow_di(this_phi, l);
  double corr=cov/var;

  /* variables required from Genz's routines for computing 
     the bivariate Gaussian probability functions */ 
  double lim_inf[2]={0.0, 0.0};
  double lim_sup[2]={0.0, 0.0};
  int infin[2]={0, 0};

  build_limits(data[t], means[t], sd, 
	       &lim_inf[0], &lim_sup[0], &infin[0]);
    
  build_limits(data[t-l], means[t-l], sd, 
	       &lim_inf[1], &lim_sup[1], &infin[1]);

 (*result)=F77_CALL(bvnmvn)(lim_inf, lim_sup, infin, &corr);

 /* if the probability is not too small... */
 if ( (*result)<SMALL)
   (*result)=SMALL;
}

/* exact score and hessian*/
static void score_info_logPair_period()
{

  int t, i, j, k;
  double prob_pair=0.0;
  
  /* important! initialise to zero */
  for(i=0; i<Dim.num_param; i++)
    score_period[i]=0.0;
  for(i=0; i<Dim.num_param; i++)
    for(j=0; j<Dim.num_param; j++)
      info_period[i+j*Dim.num_param]=0.0;

  for (t=(start_index+1); t<end_index; t++){

    /* if the data at time t is not missing... */
    if(!IS_MISSING(data[t])){
      for(k=1; k<=imin2(order_pair, t-start_index); k++){
	if(!IS_MISSING(data[t-k])){
	  score_info_log_prob_pair(t, k, score_this_pair, info_this_pair, approx_hessian);
	  
	   for(i=0; i<Dim.num_param; i++)
	     score_period[i]+=score_this_pair[i];
	   for(i=0; i<Dim.num_param; i++)
	     for(j=0; j<Dim.num_param; j++)
	       info_period[i+j*Dim.num_param]+=info_this_pair[i+j*Dim.num_param];
	}
      }
    }
  }
  
  for(i=0; i<Dim.num_param; i++)
    score_subject[this_subject][i]+=score_period[i];
  for(i=0; i<Dim.num_param; i++)
    for(j=0; j<Dim.num_param; j++)
      info_subject[this_subject][i+j*Dim.num_param]+=
	info_period[i+j*Dim.num_param];
}



//1st partial derivative of CDF_2(x, y; rho) wrt x 
//CDF_2(x,y; rho):= CDF_D bivariate normal with mean vector 0 and correlation corr evaluated at x, y
double dCDF2(double x, double y, double corr){
	return dnorm(x, 0.0, 1.0, 0)*pnorm(((y-corr*x)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
}

//2nd partial derivative of CDF_2(x, y; rho) wrt x-x 
double d2CDF2_xx(double x, double y, double corr){

	double res = 0.0, 
				 sqrt_corr = sqrt(1.0-corr*corr),
				 dn = dnorm(x, 0.0, 1.0, 0);

	res = -x*dn*pnorm(((y-corr*x)/sqrt_corr), 0.0, 1.0, 1, 0) - dn*dnorm(((y-corr*x)/sqrt_corr), 0.0, 1.0, 0)*corr/sqrt_corr;

	return res;
}

//2nd partial derivative of CDF_2(x, y; rho) wrt x-y 
double d2CDF2_xy(double x, double y, double corr){

	double res = 0.0, 
				 sqrt_corr = sqrt(1.0-corr*corr),
				 dn = dnorm(x, 0.0, 1.0, 0);

	res = dn*dnorm(((y-corr*x)/sqrt_corr), 0.0, 1.0, 0)/sqrt_corr;

	return res;
}


//2nd partial derivative of CDF_2(x, y; rho) wrt x-corr 
double d2CDF2_xcorr(double x, double y, double corr){

	double res = 0.0, 
				 sqrt_corr = sqrt(1.0-corr*corr),
				 dn = dnorm(x, 0.0, 1.0, 0);

	res = dn*dnorm(((y-corr*x)/sqrt_corr), 0.0, 1.0, 0)*(1.0/(sqrt_corr*sqrt_corr))*(-x*sqrt_corr + (y-corr*x)*corr/sqrt_corr);

	return res;
}



//PDF bivariate normal with null vector of means and correlation corr
static double dnorm2(const double x, const double y, const double corr){

  double tmp=1.0-corr*corr;
  return ( 1.0/(M_2PI*sqrt(tmp))*exp(-0.5*(x*x+y*y-2.0*corr*x*y)/tmp) );

}

//derivative PDF bivariate normal with null vector of means and correlation corr
static double der_dnorm2(const double x, const double y, const double corr){

	double corr2 = corr*corr;
  double tmp=1.0-corr2;
  return -dnorm2(x, y, corr)*( x*x*corr - x*y*(1.0+corr2) + corr*(corr2+y*y-1.0) )/(tmp*tmp);
}

//function that swaps two ids
void swap_id(int *id1_out, int *id2_out, int id1_in, int id2_in){

	int id_tmp;

	if(id1_in < id2_in){
		id_tmp = id1_in;
		*id1_out = id2_in;
		*id2_out = id_tmp;
	}else{
		*id1_out = id1_in;
		*id2_out = id2_in;
	}
}


/*#if NL*/
//NL's implementation
void NL_score_info_log_prob_pair(const int position,
			      const int distance_neigh,
			      double *score,
			      double *info)
{

  int l = distance_neigh;
  int i, j;
  double this_phi = Param.phi[index_dynamic];
  double prob_this_pair = 0.0;
  
  double phi_pow_l=R_pow_di(this_phi, l);
  double phi_pow_l_minus_one=R_pow_di(this_phi, l-1);
	double phi_pow_l_minus_two;
  if(l-2>=0){
	  phi_pow_l_minus_two=R_pow_di(this_phi, l-2);
	}else{
		phi_pow_l_minus_two=0.0;
	}
  
  double var = Param.sigma2+1.0;
  double sd = sqrt(var);
  double cov = Param.sigma2+phi_pow_l;
  double corr = cov/var;

  /* derivative of var wrt phi */
  double deriv_phi_var = 0.0;
  /* derivative of cov wrt phi  */
  double deriv_phi_cov = (l*phi_pow_l_minus_one);
  /* 1st and 2nd derivatives of corr wrt phi and sigma2 */
  double deriv_phi_corr = (deriv_phi_cov*var-cov*deriv_phi_var)/R_pow_di(var,2), 
  			 deriv_2_phi_corr = (double)(l*(l-1))*phi_pow_l_minus_two/var; 
  double deriv_sigma2_corr = (var-cov)/R_pow_di(var, 2),
			   deriv_2_sigma2_corr = -2.0*(var-cov)/R_pow_di(var, 3); 
	/* 2nd derivative of corr wrt to phi-sigma2 */
  double deriv_sigma2_phi_corr = -deriv_phi_cov/R_pow_di(var, 2);

  for(i=0; i<Dim.num_param; i++)
    score[i] = 0.0;

	double tmp_info[Dim.num_param][Dim.num_param];
  for(i=0; i<Dim.num_param; i++){
	  for(j=0; j<Dim.num_param; j++){
			tmp_info[i][j] = 0.0;
	  }
	}

  prob_pair(position, l, &prob_this_pair);

	if(prob_this_pair>SMALL){

	//observed values
	// Y_ij=s, Y_ik=t	
  int s, t; 
		s = data[position];
		t = data[position-l];

	//initialize lower and upper standardized limits 
	//l_ij < Z_ij <= u_ij
		//l_ij=(alpha_{s-1}-mu_{ij})/sd
		//u_ij=(alpha_{s}-mu_{ij})/sd

	//l_ik < Z_ik <= u_ik
		//l_ik=(alpha_{t-1}-mu_{ik})/sd
		//u_ik=(alpha_{t}-mu_{ik})/sd

  double l_ij = -BIG, 
  			 u_ij = BIG, 
  			 l_ik = -BIG, 
  			 u_ik = BIG;
    
  /* split the Gaussian rectangular probability into
     four cumulative Gaussian distribution functions */
  
  /* P(Y_ij = s,Y_ik = t; corr) =   P(Z_ij <= u_ij, Z_ik <= u_ik; corr) - P(Z_ij <= l_ij, Z_ik <= u_ik; corr)
	 	 															- P(Z_ij <= u_ij, Z_ik <= l_ik; corr) + P(Z_ij <= l_ij, Z_ik <= l_ik; corr) 
	 	 														= 	Q1(u_ij,u_ik; corr) - Q2(l_ij, u_ik; corr) +
	 	 															- Q3(u_ij, l_ik; corr) + Q4(l_ij, l_ik; corr)
	 	 														= 	Q1 - Q2 - Q3 - Q4 */

	//assign values to standardized limits
	if(s >= 2 & s <= Dim.num_categ-1){
		l_ij = (Param.tau[s-2]-means[position])/sd;
		u_ij = (Param.tau[s-1]-means[position])/sd;
	}else{
		if(s == 1){
			u_ij = (Param.tau[s-1]-means[position])/sd;
		}else{
			l_ij = (Param.tau[s-2]-means[position])/sd;
		}
	}

	if(t >= 2 & t <= Dim.num_categ-1){
		l_ik = (Param.tau[t-2]-means[position-l])/sd;
		u_ik = (Param.tau[t-1]-means[position-l])/sd;
	}else{
		if(t == 1){
			u_ik = (Param.tau[t-1]-means[position-l])/sd;
		}else{
			l_ik = (Param.tau[t-2]-means[position-l])/sd;
		}
	}


	double Q = prob_this_pair;

	//1st-order partial derivatives of Q1, Q2, Q3, and Q4 wrt to 1st/2nd argument 
	double Q1_10, Q1_01,
				 Q2_10, Q2_01,
				 Q3_10, Q3_01,
				 Q4_10, Q4_01;
		 
		//derivative Q1 wrt u_ij
		Q1_10 = dCDF2(u_ij, u_ik, corr);
		//derivative Q1 wrt u_ik
		Q1_01 = dCDF2(u_ik, u_ij, corr);

		//derivative Q2 wrt l_ij
		Q2_10 = dCDF2(l_ij, u_ik, corr);
		//derivative Q2 wrt u_ik
		Q2_01 = dCDF2(u_ik, l_ij, corr);

		//derivative Q3 wrt u_ij
		Q3_10 = dCDF2(u_ij, l_ik, corr);
		//derivative Q3 wrt l_ik
		Q3_01 = dCDF2(l_ik, u_ij, corr);

		//derivative Q4 wrtl_ij
		Q4_10 = dCDF2(l_ij, l_ik, corr);
		//derivative Q4 wrtl_ik
		Q4_01 = dCDF2(l_ik, l_ij, corr);

	//2nd-order partial derivatives of 
		//Q1_10, Q1_01 
		//Q2_10, Q2_01
		//Q3_10, Q3_01
		//Q4_10, Q4_01 
	//wrt to 1st/2nd arguments 

	double Q1_20, Q1_02, Q1_11,
				 Q2_20, Q2_02, Q2_11,
				 Q3_20, Q3_02, Q3_11,
				 Q4_20, Q4_02, Q4_11;

		//2nd partial derivative of Q1 wrt u_ij-u_ij
		Q1_20 = d2CDF2_xx(u_ij, u_ik, corr);
		//2nd partial derivative of Q1 wrt u_ik-u_ik
		Q1_02 = d2CDF2_xx(u_ik, u_ij, corr);
		//2nd partial derivative of Q1 wrt u_ij-u_ik
		Q1_11 = d2CDF2_xy(u_ij, u_ik, corr);

		//2nd partial derivative of Q2 wrt l_ij-l_ij
		Q2_20 = d2CDF2_xx(l_ij, u_ik, corr);
		//2nd partial derivative of Q2 wrt u_ik-u_ik
		Q2_02 = d2CDF2_xx(u_ik, l_ij, corr);
		//2nd partial derivative of Q2 wrt l_ij-u_ik
		Q2_11 = d2CDF2_xy(l_ij, u_ik, corr);

		//2nd partial derivative of Q3 wrt u_ij-u_ij
		Q3_20 = d2CDF2_xx(u_ij, l_ik, corr);
		//2nd partial derivative of Q3 wrt u_ik-u_ik
		Q3_02 = d2CDF2_xx(l_ik, u_ij, corr);
		//2nd partial derivative of Q3 wrt l_ij-u_ik
		Q3_11 = d2CDF2_xy(u_ij, l_ik, corr);

		//2nd partial derivative of Q4 wrt l_ij-l_ij
		Q4_20 = d2CDF2_xx(l_ij, l_ik, corr);
		//2nd partial derivative of Q4 wrt l_ik-l_ik
		Q4_02 = d2CDF2_xx(l_ik, l_ij, corr);
		//2nd partial derivative of Q4 wrt l_ij-l_ik
		Q4_11 = d2CDF2_xy(l_ij, l_ik, corr);


	//2nd-order partial derivatives of 
		//Q1_10, Q1_01 
		//Q2_10, Q2_01
		//Q3_10, Q3_01
		//Q4_10, Q4_01 
	//wrt to 3rd argument 
	double Q1_101, Q1_011,
				 Q2_101, Q2_011,
				 Q3_101, Q3_011,
				 Q4_101, Q4_011;
			 
		//derivative Q1_10 wrt corr
		Q1_101 = d2CDF2_xcorr(u_ij, u_ik, corr);
		//derivative Q1_01 wrt corr
		Q1_011 = d2CDF2_xcorr(u_ik, u_ij, corr);

		//derivative Q2_10 wrt corr
		Q2_101 = d2CDF2_xcorr(l_ij, u_ik, corr);
		//derivative Q2_01 wrt corr
		Q2_011 = d2CDF2_xcorr(u_ik, l_ij, corr);

		//derivative Q3_10 wrt corr
		Q3_101 = d2CDF2_xcorr(u_ij, l_ik, corr);
		//derivative Q3_01 wrt corr
		Q3_011 = d2CDF2_xcorr(l_ik, u_ij, corr);

		//derivative Q4_10 wrt corr
		Q4_101 = d2CDF2_xcorr(l_ij, l_ik, corr);
		//derivative Q4_01 wrt corr
		Q4_011 = d2CDF2_xcorr(l_ik, l_ij, corr);


	//pdf bivariate normal
	double PDF[4];
		PDF[0] = dnorm2(u_ij, u_ik, corr);
		PDF[1] = dnorm2(l_ij, u_ik, corr);
		PDF[2] = dnorm2(u_ij, l_ik, corr);
		PDF[3] = dnorm2(l_ij, l_ik, corr);


	//1st and 2nd order parial derivatives of Q wrt 3rd argument
	double Q001 = PDF[0]-PDF[1]-PDF[2]+PDF[3];
	double Q002 = der_dnorm2(u_ij, u_ik, corr) - der_dnorm2(u_ij, l_ik, corr) - der_dnorm2(l_ij, u_ik, corr) + der_dnorm2(l_ij, l_ik, corr);

	//1st-order partial derivative of endpoints wrt regressors
	double u_ij_beta[Dim.num_regr], u_ik_beta[Dim.num_regr];
		for(i=0; i<Dim.num_regr; i++){
			u_ij_beta[i] = -design_matrix[position][i]/sd;
			u_ik_beta[i] = -design_matrix[position-l][i]/sd;
		}
		
	//2nd-order partial derivative of endpoints wrt regressors	
	double u_ij_s2_beta[Dim.num_regr], u_ik_s2_beta[Dim.num_regr];
		for(i=0; i<Dim.num_regr; i++){
			u_ij_s2_beta[i] = 0.5*design_matrix[position][i]/(sd*var);
			u_ik_s2_beta[i] = 0.5*design_matrix[position-l][i]/(sd*var);
		}


	//////////////////////////
	//Computation of the score
	//////////////////////////
	
	//1st-order partial derivatives wrt thresholds
		if(s >= 2 & s <= Dim.num_categ-1){
			//wrt alpha_s, i.e, Param.tau[s-2]
			score[s-2] += (Q1_10 - Q3_10)/sd; 
		}
		if(s >= 3){	
			//wrt alpha_{s-1}, i.e,  Param.tau[s-3]
			score[s-3] += (-Q2_10 + Q4_10)/sd;
		}

		if(t >= 2 & t <= Dim.num_categ-1){
			//wrt alpha_t, i.e., Param.tau[t-2]
				score[t-2] += (Q1_01 - Q2_01)/sd;
		}
		if(t >= 3){	
			//wrt alpha_{t-1}, i.e.,  Param.tau[t-3]
				score[t-3] += (-Q3_01 + Q4_01)/sd;
		}

	//1st-order partial derivatives wrt regressors/random effect/autocorrelation
		//autocorreation
	  score[(Dim.num_categ-2)+Dim.num_regr+index_dynamic] += Q001*deriv_phi_corr;

		//common contribution to score wrt random effect
		if(Dim.is_random_effect)
			score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr] += Q001*deriv_sigma2_corr;

		if(s <= Dim.num_categ-1){ //u_ij is finite
			//regressors
			for(i=0; i<Dim.num_regr; i++){
				score[i+Dim.num_categ-2] += (Q1_10 - Q3_10)*u_ij_beta[i];
			}
			//random effect
			if(Dim.is_random_effect)
				score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr] += -0.5*(Q1_10-Q3_10)*u_ij/var;
		}
		if(s >= 2){	//l_ik is finite
			//regressors
			for(i=0; i<Dim.num_regr; i++){
				score[i+Dim.num_categ-2] += (-Q2_10 + Q4_10)*u_ij_beta[i];
			}
			//random effect
			if(Dim.is_random_effect)
				score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr] += -0.5*(-Q2_10+Q4_10)*l_ij/var;
		}
		if(t <= Dim.num_categ-1){ //u_ik is finite
			//regressors
			for(i=0; i<Dim.num_regr; i++){
				score[i+Dim.num_categ-2] += (Q1_01 - Q2_01)*u_ik_beta[i];
			}
			//random effect
			if(Dim.is_random_effect)
				score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr] += -0.5*(Q1_01-Q2_01)*u_ik/var;
		}
		if(t >= 2){	//l_ik is finite
			//regressors
			for(i=0; i<Dim.num_regr; i++){
				score[i+Dim.num_categ-2] += (-Q3_01 + Q4_01)*u_ik_beta[i];
			}
			//random effect
			if(Dim.is_random_effect)
				score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr] += -0.5*(-Q3_01+Q4_01)*l_ik/var;
		}

		for(i=0; i<Dim.num_param; i++)
			score[i] /= Q;




	//////////////////////////
	//Computation of the hessian
	//////////////////////////
	int id_r, id_c;

	//derivatives of u_ij, l_ij, u_ik, l_ik wrt random effect
	double u_ij_s2 = -0.5*u_ij/var,
				 l_ij_s2 = -0.5*l_ij/var,
				 u_ik_s2 = -0.5*u_ik/var,
				 l_ik_s2 = -0.5*l_ik/var;  

	//derivatives of u_ij, l_ij, u_ik, l_ik wrt random effect-random effect
	double u_ij_s22 = 0.75*u_ij/(var*var),
				 l_ij_s22 = 0.75*l_ij/(var*var),
				 u_ik_s22 = 0.75*u_ik/(var*var),
				 l_ik_s22 = 0.75*l_ik/(var*var);  


	//2nd-order partial derivative wrt to autocorrelation-autocorrelation
	id_r = (Dim.num_categ-2)+Dim.num_regr+index_dynamic;
	tmp_info[id_r][id_r] += -Q001*Q001*deriv_phi_corr*deriv_phi_corr/(Q*Q) + Q002*deriv_phi_corr*deriv_phi_corr/Q + Q001*deriv_2_phi_corr/Q;


	//all derivatives that depend on random effect need to be placed right here (within "if(Dim.is_random_effect)")

	if(Dim.is_random_effect){
	
		id_r = (Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr;
		id_c = (Dim.num_categ-2)+Dim.num_regr+index_dynamic;
		//random effect-random effect
		//part without conditions on s, t
		tmp_info[id_r][id_r] += -(Q001*deriv_sigma2_corr)*(Q001*deriv_sigma2_corr)/(Q*Q) + Q002*deriv_sigma2_corr*deriv_sigma2_corr/Q 
														+ Q001*deriv_2_sigma2_corr/Q;
		//
		tmp_info[id_r][id_r] += -(Q1_10-Q3_10)*u_ij_s2*Q001*deriv_sigma2_corr/(Q*Q) + (Q1_101-Q3_101)*deriv_sigma2_corr*u_ij_s2/Q;
		tmp_info[id_r][id_r] += -(-Q2_10+Q4_10)*l_ij_s2*Q001*deriv_sigma2_corr/(Q*Q) + (-Q2_101+Q4_101)*deriv_sigma2_corr*l_ij_s2/Q;
		tmp_info[id_r][id_r] += -(Q1_01-Q2_01)*u_ik_s2*Q001*deriv_sigma2_corr/(Q*Q) + (Q1_011-Q2_011)*deriv_sigma2_corr*u_ik_s2/Q;
		tmp_info[id_r][id_r] += -(-Q3_01+Q4_01)*l_ik_s2*Q001*deriv_sigma2_corr/(Q*Q) + (-Q3_011+Q4_011)*deriv_sigma2_corr*l_ik_s2/Q;
		//

		//autocorrelation-random effect
		//part without conditions on s, t (i.e., derivatives wrt correlation)
		tmp_info[id_r][id_c] += - Q001*Q001*deriv_phi_corr*deriv_sigma2_corr/(Q*Q) + Q002*deriv_phi_corr*deriv_sigma2_corr/Q 
													  + Q001*deriv_sigma2_phi_corr/Q;
	
		if( s <= Dim.num_categ-1 ){
		//autocorrelation-random effect
		tmp_info[id_r][id_c] += -(Q1_10-Q3_10)*u_ij_s2*Q001*deriv_phi_corr/(Q*Q) + (Q1_101-Q3_101)*u_ij_s2*deriv_phi_corr/Q;

		//random effect-random effect
/*		tmp_info[id_r][id_r] += -(Q1_10-Q3_10)*u_ij_s2*Q001*deriv_sigma2_corr/(Q*Q) + (Q1_101-Q3_101)*deriv_sigma2_corr*u_ij_s2/Q;*/
		tmp_info[id_r][id_r] += -( Q001*deriv_sigma2_corr + (Q1_10-Q3_10)*u_ij_s2 )*(Q1_10-Q3_10)*u_ij_s2/(Q*Q) 
													 + (Q1_20-Q3_20)*u_ij_s2*u_ij_s2/Q + (Q1_101-Q3_101)*u_ij_s2*deriv_sigma2_corr/Q + (Q1_10-Q3_10)*u_ij_s22/Q;

		//regressors-random effect 
		for(i=0; i<Dim.num_regr; i++){
			tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q001*deriv_sigma2_corr + (Q1_10-Q3_10)*u_ij_s2)*(Q1_10-Q3_10)*u_ij_beta[i]/(Q*Q) 
									+ ((Q1_101-Q3_101)*deriv_sigma2_corr+(Q1_20-Q3_20)*u_ij_s2)*u_ij_beta[i]/Q + (Q1_10-Q3_10)*u_ij_s2_beta[i]/Q;
		}

			if( s >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q2_10+Q4_10)*l_ij_s2*(Q1_10-Q3_10)*u_ij_s2/(Q*Q);
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q2_10+Q4_10)*l_ij_s2*(Q1_10-Q3_10)*u_ij_beta[i]/(Q*Q);
				}
			}

			if( t <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_01-Q2_01)*u_ik_s2*(Q1_10-Q3_10)*u_ij_s2/(Q*Q) + Q1_11*u_ik_s2*u_ij_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_01-Q2_01)*u_ik_s2*(Q1_10-Q3_10)*u_ij_beta[i]/(Q*Q) + Q1_11*u_ik_s2*u_ij_beta[i]/Q;
				}
			}

			if( t >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q3_01+Q4_01)*l_ik_s2*(Q1_10-Q3_10)*u_ij_s2/(Q*Q) - Q3_11*l_ik_s2*u_ij_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q3_01+Q4_01)*l_ik_s2*(Q1_10-Q3_10)*u_ij_beta[i]/(Q*Q) -Q3_11*l_ik_s2*u_ij_beta[i]/Q;
				}
			}
		}

		if( s >= 2 ){
			//autocorrelation-random effect
			tmp_info[id_r][id_c] += -(-Q2_10+Q4_10)*l_ij_s2*Q001*deriv_phi_corr/(Q*Q) + (-Q2_101+Q4_101)*l_ij_s2*deriv_phi_corr/Q;

			//random effect-random effect
/*			tmp_info[id_r][id_r] += -(-Q2_10+Q4_10)*l_ij_s2*Q001*deriv_sigma2_corr/(Q*Q) + (-Q2_101+Q4_101)*deriv_sigma2_corr*l_ij_s2/Q;*/
			tmp_info[id_r][id_r] += -( Q001*deriv_sigma2_corr + (-Q2_10+Q4_10)*l_ij_s2 )*(-Q2_10+Q4_10)*l_ij_s2/(Q*Q) 
														 + (-Q2_20+Q4_20)*l_ij_s2*l_ij_s2/Q + (-Q2_101+Q4_101)*l_ij_s2*deriv_sigma2_corr/Q + (-Q2_10+Q4_10)*l_ij_s22/Q;

			//regressors-random effect  
			for(i=0; i<Dim.num_regr; i++){
				tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q001*deriv_sigma2_corr + (-Q2_10+Q4_10)*l_ij_s2)*(-Q2_10+Q4_10)*u_ij_beta[i]/(Q*Q) 
										+ ((-Q2_101+Q4_101)*deriv_sigma2_corr+(-Q2_20+Q4_20)*l_ij_s2)*u_ij_beta[i]/Q + (-Q2_10+Q4_10)*u_ij_s2_beta[i]/Q;
			}

			if( s <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_10-Q3_10)*u_ij_s2*(-Q2_10+Q4_10)*l_ij_s2/(Q*Q);
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_10-Q3_10)*u_ij_s2*(-Q2_10+Q4_10)*u_ij_beta[i]/(Q*Q);
				}
			}

			if( t <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_01-Q2_01)*u_ik_s2*(-Q2_10+Q4_10)*l_ij_s2/(Q*Q) - Q2_11*u_ik_s2*l_ij_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_01-Q2_01)*u_ik_s2*(-Q2_10+Q4_10)*u_ij_beta[i]/(Q*Q) - Q2_11*u_ik_s2*u_ij_beta[i]/Q;
				}
			}

			if( t >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q3_01+Q4_01)*l_ik_s2*(-Q2_10+Q4_10)*l_ij_s2/(Q*Q) + Q4_11*l_ik_s2*l_ij_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q3_01+Q4_01)*l_ik_s2*(-Q2_10+Q4_10)*u_ij_beta[i]/(Q*Q) + Q4_11*l_ik_s2*u_ij_beta[i]/Q;
				}
			}
		}
		
		if( t <= Dim.num_categ-1 ){
			//autocorrelation-random effect
			tmp_info[id_r][id_c] += -(Q1_01-Q2_01)*u_ik_s2*Q001*deriv_phi_corr/(Q*Q) + (Q1_011-Q2_011)*u_ik_s2*deriv_phi_corr/Q;

			//random effect-random effect
/*			tmp_info[id_r][id_r] += -(Q1_01-Q2_01)*u_ik_s2*Q001*deriv_sigma2_corr/(Q*Q) + (Q1_011-Q2_011)*deriv_sigma2_corr*u_ik_s2/Q;*/
			tmp_info[id_r][id_r] += -( Q001*deriv_sigma2_corr + (Q1_01-Q2_01)*u_ik_s2 )*(Q1_01-Q2_01)*u_ik_s2/(Q*Q) + (Q1_02-Q2_02)*u_ik_s2*u_ik_s2/Q + (Q1_011-Q2_011)*u_ik_s2*deriv_sigma2_corr/Q + (Q1_01-Q2_01)*u_ik_s22/Q;

			//regressors-random effect  
			for(i=0; i<Dim.num_regr; i++){
				tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q001*deriv_sigma2_corr + (Q1_01-Q2_01)*u_ik_s2)*(Q1_01-Q2_01)*u_ik_beta[i]/(Q*Q) 
										+ ((Q1_011-Q2_011)*deriv_sigma2_corr+(Q1_02-Q2_02)*u_ik_s2)*u_ik_beta[i]/Q + (Q1_01-Q2_01)*u_ik_s2_beta[i]/Q;
			}

			if( t >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q3_01+Q4_01)*l_ik_s2*(Q1_01-Q2_01)*u_ik_s2/(Q*Q);
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q3_01+Q4_01)*l_ik_s2*(Q1_01-Q2_01)*u_ik_beta[i]/(Q*Q);
				}
			}

			if( s <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_10-Q3_10)*u_ij_s2*(Q1_01-Q2_01)*u_ik_s2/(Q*Q) + Q1_11*u_ij_s2*u_ik_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_10-Q3_10)*u_ij_s2*(Q1_01-Q2_01)*u_ik_beta[i]/(Q*Q) + Q1_11*u_ij_s2*u_ik_beta[i]/Q;
				}
			}

			if( s >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q2_10+Q4_10)*l_ij_s2*(Q1_01-Q2_01)*u_ik_s2/(Q*Q) - Q2_11*l_ij_s2*u_ik_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q2_10+Q4_10)*l_ij_s2*(Q1_01-Q2_01)*u_ik_beta[i]/(Q*Q) - Q2_11*l_ij_s2*u_ik_beta[i]/Q;
				}
			}
		}

		if( t >= 2 ){
			//autocorrelation-random effect
			tmp_info[id_r][id_c] += -(-Q3_01+Q4_01)*l_ik_s2*Q001*deriv_phi_corr/(Q*Q) + (-Q3_011+Q4_011)*l_ik_s2*deriv_phi_corr/Q;

			//random effect-random effect
/*			tmp_info[id_r][id_r] += -(-Q3_01+Q4_01)*l_ik_s2*Q001*deriv_sigma2_corr/(Q*Q) + (-Q3_011+Q4_011)*deriv_sigma2_corr*l_ik_s2/Q;*/
			tmp_info[id_r][id_r] += -( Q001*deriv_sigma2_corr + (-Q3_01+Q4_01)*l_ik_s2 )*(-Q3_01+Q4_01)*l_ik_s2/(Q*Q) + (-Q3_02+Q4_02)*l_ik_s2*l_ik_s2/Q + (-Q3_011+Q4_011)*l_ik_s2*deriv_sigma2_corr/Q + (-Q3_01+Q4_01)*l_ik_s22/Q;

			//regressors-random effect 
			for(i=0; i<Dim.num_regr; i++){
				tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q001*deriv_sigma2_corr + (-Q3_01+Q4_01)*l_ik_s2)*(-Q3_01+Q4_01)*u_ik_beta[i]/(Q*Q) + ((-Q3_011+Q4_011)*deriv_sigma2_corr+(-Q3_02+Q4_02)*l_ik_s2)*u_ik_beta[i]/Q + (-Q3_01+Q4_01)*u_ik_s2_beta[i]/Q;
			}

			if( t <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_01-Q2_01)*u_ik_s2*(-Q3_01+Q4_01)*l_ik_s2/(Q*Q);
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_01-Q2_01)*u_ik_s2*(-Q3_01+Q4_01)*u_ik_beta[i]/(Q*Q);
				}
			}

			if( s <= Dim.num_categ-1 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(Q1_10-Q3_10)*u_ij_s2*(-Q3_01+Q4_01)*l_ik_s2/(Q*Q) - Q3_11*u_ij_s2*l_ik_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_10-Q3_10)*u_ij_s2*(-Q3_01+Q4_01)*u_ik_beta[i]/(Q*Q) - Q3_11*u_ij_s2*u_ik_beta[i]/Q;
				}
			}

			if( s >= 2 ){
				//random effect-random effect
				tmp_info[id_r][id_r] += -(-Q2_10+Q4_10)*l_ij_s2*(-Q3_01+Q4_01)*l_ik_s2/(Q*Q) + Q4_11*l_ij_s2*l_ik_s2/Q;
				//regressors-random effect 
				for(i=0; i<Dim.num_regr; i++){
					tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q2_10+Q4_10)*l_ij_s2*(-Q3_01+Q4_01)*u_ik_beta[i]/(Q*Q) + Q4_11*l_ij_s2*u_ik_beta[i]/Q;
				}
			}
		}
	}


	////////////////////////////
	//autocorrelation-regressors
	////////////////////////////
	id_r = (Dim.num_categ-2)+Dim.num_regr+index_dynamic;
	if( s <= Dim.num_categ-1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_10-Q3_10)*u_ij_beta[i]*Q001*deriv_phi_corr/(Q*Q) + (Q1_101-Q3_101)*u_ij_beta[i]*deriv_phi_corr/Q;
			}
	}
	if( s >= 2 ){ 
		for(i=0; i<Dim.num_regr; i++){
			tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q2_10+Q4_10)*u_ij_beta[i]*Q001*deriv_phi_corr/(Q*Q) + (-Q2_101+Q4_101)*u_ij_beta[i]*deriv_phi_corr/Q;
			}
	}

	if( t <= Dim.num_categ-1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			tmp_info[id_r][(Dim.num_categ-2)+i] += -(Q1_01-Q2_01)*u_ik_beta[i]*Q001*deriv_phi_corr/(Q*Q) + (Q1_011-Q2_011)*u_ik_beta[i]*deriv_phi_corr/Q;
			}
	}
	if( t >= 2 ){ 
		for(i=0; i<Dim.num_regr; i++){
			tmp_info[id_r][(Dim.num_categ-2)+i] += -(-Q3_01+Q4_01)*u_ik_beta[i]*Q001*deriv_phi_corr/(Q*Q) + (-Q3_011+Q4_011)*u_ik_beta[i]*deriv_phi_corr/Q;
			}
	}
	
	//////////////////////////////////////////
	//autocorrelation/random effect-thresholds
	//////////////////////////////////////////
	if(s >= 2 & s <= Dim.num_categ-1){
		//autocorrelation-thresholds
		swap_id(&id_r, &id_c, s-2, (Dim.num_categ-2)+Dim.num_regr+index_dynamic);	
		tmp_info[id_r][id_c] += (-Q001*(Q1_10 - Q3_10)/(Q*Q) + (Q1_101 - Q3_101)/Q )*deriv_phi_corr/sd; 

		//random effects-thresholds
		if(Dim.is_random_effect){
			swap_id(&id_r, &id_c, s-2, (Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr);	
			tmp_info[id_r][id_c] += -(Q001*deriv_sigma2_corr + (Q1_10-Q3_10)*u_ij_s2 + (-Q2_10 + Q4_10)*l_ij_s2)*(Q1_10 - Q3_10)/(sd*Q*Q) - 0.5*(Q1_10 - Q3_10)/(sd*var*Q) + u_ij_s2*(Q1_20-Q3_20)/(sd*Q) + deriv_sigma2_corr*(Q1_101-Q3_101)/(sd*Q);

			if( t <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_01 - Q2_01)*u_ik_s2*(Q1_10 - Q3_10)/(sd*Q*Q) + u_ik_s2*Q1_11/(sd*Q);

			if( t >= 2 )
				tmp_info[id_r][id_c] += -(-Q3_01 + Q4_01)*l_ik_s2*(Q1_10 - Q3_10)/(sd*Q*Q) - l_ik_s2*Q3_11/(sd*Q);
		}
	}
	if(s >= 3){	
		//autocorrelation-thresholds
		swap_id(&id_r, &id_c, s-3, (Dim.num_categ-2)+Dim.num_regr+index_dynamic);	
		tmp_info[id_r][id_c] += (-Q001*(-Q2_10 + Q4_10)/(Q*Q) + (-Q2_101 + Q4_101)/Q )*deriv_phi_corr/sd;

		//random effects-thresholds
		if(Dim.is_random_effect){
			swap_id(&id_r, &id_c, s-3, (Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr);	
			tmp_info[id_r][id_c] += -(Q001*deriv_sigma2_corr + (-Q2_10 + Q4_10)*l_ij_s2)*(-Q2_10 + Q4_10)/(sd*Q*Q) - 0.5*(-Q2_10 + Q4_10)/(sd*var*Q) + l_ij_s2*(-Q2_20 + Q4_20)/(sd*Q) + deriv_sigma2_corr*(-Q2_101 + Q4_101)/(sd*Q);			
			
			if( s <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_10 - Q3_10)*u_ij_s2*(-Q2_10 + Q4_10)/(sd*Q*Q);

			if( t <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_01 - Q2_01)*u_ik_s2*(-Q2_10 + Q4_10)/(sd*Q*Q) - u_ik_s2*Q2_11/(sd*Q);

			if( t >= 2 )
				tmp_info[id_r][id_c] += -(-Q3_01 + Q4_01)*l_ik_s2*(-Q2_10 + Q4_10)/(sd*Q*Q) + l_ik_s2*Q4_11/(sd*Q);
		}
	}
	if(t >= 2 & t <= Dim.num_categ-1){
		//autocorrelation-thresholds
		swap_id(&id_r, &id_c, t-2, (Dim.num_categ-2)+Dim.num_regr+index_dynamic);	
		tmp_info[id_r][id_c] += (-Q001*(Q1_01 - Q2_01)/(Q*Q) + (Q1_011 - Q2_011)/Q )*deriv_phi_corr/sd; 

		//random effects-thresholds
		if(Dim.is_random_effect){
			swap_id(&id_r, &id_c, t-2, (Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr);	
			tmp_info[id_r][id_c] += -(Q001*deriv_sigma2_corr + (Q1_01-Q2_01)*u_ik_s2 + (-Q3_01 + Q4_01)*l_ik_s2)*(Q1_01 - Q2_01)/(sd*Q*Q) - 0.5*(Q1_01 - Q2_01)/(sd*var*Q) + u_ik_s2*(Q1_02-Q2_02)/(sd*Q) + deriv_sigma2_corr*(Q1_011-Q2_011)/(sd*Q);

			if( s <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_10 - Q3_10)*u_ij_s2*(Q1_01 - Q2_01)/(sd*Q*Q) + u_ij_s2*Q1_11/(sd*Q);

			if( s >= 2 )
				tmp_info[id_r][id_c] += -(-Q2_10 + Q4_10)*l_ij_s2*(Q1_01 - Q2_01)/(sd*Q*Q) - l_ij_s2*Q2_11/(sd*Q);
		}
	}
	if(t >= 3){	
		//autocorrelation-thresholds
		swap_id(&id_r, &id_c, t-3, (Dim.num_categ-2)+Dim.num_regr+index_dynamic);	
		tmp_info[id_r][id_c] += (-Q001*(-Q3_01 + Q4_01)/(Q*Q) + (-Q3_011 + Q4_011)/Q )*deriv_phi_corr/sd; 

		//random effects-thresholds
		if(Dim.is_random_effect){
			swap_id(&id_r, &id_c, t-3, (Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr);	
			tmp_info[id_r][id_c] += -(Q001*deriv_sigma2_corr + (-Q3_01+Q4_01)*l_ik_s2)*(-Q3_01 + Q4_01)/(sd*Q*Q) - 0.5*(-Q3_01 + Q4_01)/(sd*var*Q) + l_ik_s2*(-Q3_02+Q4_02)/(sd*Q) + deriv_sigma2_corr*(-Q3_011+Q4_011)/(sd*Q);

			if( t <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_01 - Q2_01)*u_ik_s2*(-Q3_01 + Q4_01)/(sd*Q*Q);

			if( s <= Dim.num_categ - 1 )
				tmp_info[id_r][id_c] += -(Q1_10 - Q3_10)*u_ij_s2*(-Q3_01 + Q4_01)/(sd*Q*Q) - u_ij_s2*Q3_11/(sd*Q);

			if( s >= 2 )
				tmp_info[id_r][id_c] += -(-Q2_10 + Q4_10)*l_ij_s2*(-Q3_01 + Q4_01)/(sd*Q*Q) + l_ij_s2*Q4_11/(sd*Q);
		}
	}

	///////////////////////
	//thresholds-thresholds
	///////////////////////
	if(s >= 2 & s <= Dim.num_categ-1){
		//wrt alpha_s alpha_s
		tmp_info[s-2][s-2] += (-(Q1_10 - Q3_10)*(Q1_10 - Q3_10)/(Q*Q) + (Q1_20-Q3_20)/Q)/var; 

		if( s >= 3 ){
			swap_id(&id_r, &id_c, s-2, s-3);
			tmp_info[id_r][id_c] += -(-Q2_10 + Q4_10)*(Q1_10 - Q3_10)/(Q*Q*var); 
		}

		//wrt alpha_s alpha_t
		if( t >= 2 & t <= Dim.num_categ-1 ){
			swap_id(&id_r, &id_c, s-2, t-2);
			tmp_info[id_r][id_c] += (-(Q1_01 - Q2_01)*(Q1_10 - Q3_10)/(Q*Q) + Q1_11/Q)/var; 
		}
		//wrt alpha_s alpha_{t-1}
		if( t >= 3 ){
			swap_id(&id_r, &id_c, s-2, t-3);
			tmp_info[id_r][id_c] += (-(-Q3_01 + Q4_01)*(Q1_10 - Q3_10)/(Q*Q) - Q3_11/Q)/var; 
		}
	}

	if( s >= 3 ){
		//wrt alpha_s alpha_{s-1}
		tmp_info[s-3][s-3] += (-(-Q2_10 + Q4_10)*(-Q2_10 + Q4_10)/(Q*Q) + (-Q2_20 + Q4_20)/Q)/var; 

		if( s <= Dim.num_categ-1 ){
			swap_id(&id_r, &id_c, s-2, s-3);
			tmp_info[id_r][id_c] += (-(Q1_10 - Q3_10)*(-Q2_10 + Q4_10)/(Q*Q))/var; 
		}

		//wrt alpha_{s-1} alpha_t
		if( t >= 2 & t <= Dim.num_categ-1 ){
			swap_id(&id_r, &id_c, s-3, t-2);
			tmp_info[id_r][id_c] += (-(Q1_01 - Q2_01)*(-Q2_10 + Q4_10)/(Q*Q) - Q2_11/Q)/var; 
		}
		//wrt alpha_s alpha_{t-1}
		if( t >= 3 ){
			swap_id(&id_r, &id_c, s-3, t-3);
			tmp_info[id_r][id_c] += (-(-Q3_01 + Q4_01)*(-Q2_10 + Q4_10)/(Q*Q) + Q4_11/Q)/var; 
		}
	}


	if(t >= 2 & t <= Dim.num_categ-1){
		//wrt alpha_t alpha_t
		tmp_info[t-2][t-2] += (-(Q1_01 - Q2_01)*(Q1_01 - Q2_01)/(Q*Q) + (Q1_02 - Q2_02)/Q)/var; 

		if( t >= 3 ){
			swap_id(&id_r, &id_c, t-2, t-3);
			tmp_info[id_r][id_c] += (-(-Q3_01 + Q4_01)*(Q1_01 - Q2_01)/(Q*Q))/var; 
		}

		//wrt alpha_t alpha_s
		if(s >= 2 & s<= Dim.num_categ-1){
			swap_id(&id_r, &id_c, t-2, s-2);
			tmp_info[id_r][id_c] += (-(Q1_10 - Q3_10)*(Q1_01 - Q2_01)/(Q*Q) + Q1_11/Q)/var; 
		}
		//wrt alpha_t alpha_{s-1}
		if( s >= 3 ){
			swap_id(&id_r, &id_c, t-2, s-3);
			tmp_info[id_r][id_c] += (-(-Q2_10 + Q4_10)*(Q1_01 - Q2_01)/(Q*Q) - Q2_11/Q)/var; 
		}
	}

	if( t >= 3 ){
		//wrt alpha_{t-1} alpha_{t-1}
		tmp_info[t-3][t-3] += (-(-Q3_01 + Q4_01)*(-Q3_01 + Q4_01)/(Q*Q) + (-Q3_02 + Q4_02)/Q)/var; 

		if( t <= Dim.num_categ-1 ){
			swap_id(&id_r, &id_c, t-2, t-3);
			tmp_info[id_r][id_c] += (-(-Q3_01 + Q4_01)*(Q1_01 - Q2_01)/(Q*Q))/var; 
		}

		//wrt alpha_{t-1} alpha_{s}
		if( s >= 2 & s <= Dim.num_categ-1 ){
			swap_id(&id_r, &id_c, t-3, s-2);
			tmp_info[id_r][id_c] += (-(Q1_10 - Q3_10)*(-Q3_01 + Q4_01)/(Q*Q) - Q3_11/Q)/var; 
		}
		//wrt alpha_{t-1} alpha_{s-1}
		if( s >= 3 ){
			swap_id(&id_r, &id_c, t-3, s-3);
			tmp_info[id_r][id_c] += (-(-Q2_10 + Q4_10)*(-Q3_01 + Q4_01)/(Q*Q) + Q4_11/Q)/var; 
		}
	}


	///////////////////////
	//thresholds-regressors
	///////////////////////
	if( s >= 2 & s <= Dim.num_categ-1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-2);	
			tmp_info[id_r][id_c] += -(Q1_10-Q2_10-Q3_10+Q4_10)*(Q1_10-Q3_10)*u_ij_beta[i]/(sd*Q*Q) + (Q1_20-Q3_20)*u_ij_beta[i]/(sd*Q);
		}
		
		if( t <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-2);	
				tmp_info[id_r][id_c] += -(Q1_01-Q2_01)*(Q1_10-Q3_10)*u_ik_beta[i]/(sd*Q*Q) + Q1_11*u_ik_beta[i]/(sd*Q);
			}
		}
		
		if( t >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-2);	
				tmp_info[id_r][id_c] += -(-Q3_01+Q4_01)*(Q1_10-Q3_10)*u_ik_beta[i]/(sd*Q*Q) - Q3_11*u_ik_beta[i]/(sd*Q);
			}
		}
	}

	if( s >= 3 ){ 
		for(i=0; i<Dim.num_regr; i++){
			swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-3);	
			tmp_info[id_r][id_c] += -(-Q2_10+Q4_10)*(-Q2_10+Q4_10)*u_ij_beta[i]/(sd*Q*Q) + (-Q2_20+Q4_20)*u_ij_beta[i]/(sd*Q);
		}
		
		if( s <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-3);	
				tmp_info[id_r][id_c] += -(Q1_10-Q3_10)*(-Q2_10+Q4_10)*u_ij_beta[i]/(sd*Q*Q);
			}
		}

		if( t <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-3);	
				tmp_info[id_r][id_c] += -(Q1_01-Q2_01)*(-Q2_10+Q4_10)*u_ik_beta[i]/(sd*Q*Q) - Q2_11*u_ik_beta[i]/(sd*Q);
			}
		}
		
		if( t >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, s-3);	
				tmp_info[id_r][id_c] += -(-Q3_01+Q4_01)*(-Q2_10+Q4_10)*u_ik_beta[i]/(sd*Q*Q) + Q4_11*u_ik_beta[i]/(sd*Q);
			}
		}
	}

	if( t >= 2 & t <= Dim.num_categ-1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-2);	
			tmp_info[id_r][id_c] += -(Q1_01-Q2_01-Q3_01+Q4_01)*(Q1_01-Q2_01)*u_ik_beta[i]/(sd*Q*Q) + (Q1_02-Q2_02)*u_ik_beta[i]/(sd*Q);
		}
		
		if( s <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-2);	
				tmp_info[id_r][id_c] += -(Q1_10-Q3_10)*(Q1_01-Q2_01)*u_ij_beta[i]/(sd*Q*Q) + Q1_11*u_ij_beta[i]/(sd*Q);
			}
		}
		
		if( s >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-2);	
				tmp_info[id_r][id_c] += -(-Q2_10+Q4_10)*(Q1_01-Q2_01)*u_ij_beta[i]/(sd*Q*Q) - Q2_11*u_ij_beta[i]/(sd*Q);
			}
		}
	}

	if( t >= 3 ){ 
		for(i=0; i<Dim.num_regr; i++){
			swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-3);	
			tmp_info[id_r][id_c] += -(-Q3_01+Q4_01)*(-Q3_01+Q4_01)*u_ik_beta[i]/(sd*Q*Q) + (-Q3_02+Q4_02)*u_ik_beta[i]/(sd*Q);
		}
		
		if( t <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-3);	
				tmp_info[id_r][id_c] += -(Q1_01-Q2_01)*(-Q3_01+Q4_01)*u_ik_beta[i]/(sd*Q*Q);
				}
		}

		if( s <= Dim.num_categ-1 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-3);	
				tmp_info[id_r][id_c] += -(Q1_10-Q3_10)*(-Q3_01+Q4_01)*u_ij_beta[i]/(sd*Q*Q) - Q3_11*u_ij_beta[i]/(sd*Q);
			}
		}
		
		if( s >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				swap_id(&id_r, &id_c, (Dim.num_categ-2)+i, t-3);	
				tmp_info[id_r][id_c] += -(-Q2_10+Q4_10)*(-Q3_01+Q4_01)*u_ij_beta[i]/(sd*Q*Q) + Q4_11*u_ij_beta[i]/(sd*Q);
			}
		}
	}

	///////////////////////
	//regressors-regressors
	///////////////////////
	id_r = Dim.num_categ-2;
	if( s <= Dim.num_categ - 1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			for(j=i; j<Dim.num_regr; j++){
				tmp_info[id_r+j][id_r+i] += -(Q1_10 - Q3_10)*u_ij_beta[j]*(Q1_10 - Q3_10)*u_ij_beta[i]/(Q*Q) + (Q1_20 - Q3_20)*u_ij_beta[j]*u_ij_beta[i]/Q;
			}
		}

		if( s >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q2_10 + Q4_10)*u_ij_beta[j]*(Q1_10 - Q3_10)*u_ij_beta[i]/(Q*Q);
				}
			}
		}

		if( t <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_01 - Q2_01)*u_ik_beta[j]*(Q1_10 - Q3_10)*u_ij_beta[i]/(Q*Q) + Q1_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}

		if( t >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q3_01 + Q4_01)*u_ik_beta[j]*(Q1_10 - Q3_10)*u_ij_beta[i]/(Q*Q) - Q3_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}
	}
	

	if( s >= 2 ){ 
		for(i=0; i<Dim.num_regr; i++){
			for(j=i; j<Dim.num_regr; j++){
				tmp_info[id_r+j][id_r+i] += -(-Q2_10 + Q4_10)*u_ij_beta[j]*(-Q2_10 + Q4_10)*u_ij_beta[i]/(Q*Q) + (-Q2_20 + Q4_20)*u_ij_beta[j]*u_ij_beta[i]/Q;
			}
		}

		if( s <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_10 - Q3_10)*u_ij_beta[j]*(-Q2_10 + Q4_10)*u_ij_beta[i]/(Q*Q);
				}
			}
		}

		if( t <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_01 - Q2_01)*u_ik_beta[j]*(-Q2_10 + Q4_10)*u_ij_beta[i]/(Q*Q) - Q2_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}

		if( t >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q3_01 + Q4_01)*u_ik_beta[j]*(-Q2_10 + Q4_10)*u_ij_beta[i]/(Q*Q) + Q4_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}
	}


	if( t <= Dim.num_categ - 1 ){ 
		for(i=0; i<Dim.num_regr; i++){
			for(j=i; j<Dim.num_regr; j++){
				tmp_info[id_r+j][id_r+i] += -(Q1_01 - Q2_01)*u_ik_beta[j]*(Q1_01 - Q2_01)*u_ik_beta[i]/(Q*Q) + (Q1_02 - Q2_02)*u_ik_beta[j]*u_ik_beta[i]/Q;
			}
		}

		if( t >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q3_01 + Q4_01)*u_ik_beta[j]*(Q1_01 - Q2_01)*u_ik_beta[i]/(Q*Q);
				}
			}
		}

		if( s <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_10 - Q3_10)*u_ij_beta[j]*(Q1_01 - Q2_01)*u_ik_beta[i]/(Q*Q) + Q1_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}

		if( s >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q2_10 + Q4_10)*u_ij_beta[j]*(Q1_01 - Q2_01)*u_ik_beta[i]/(Q*Q) - Q2_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}
	}


	if( t  >= 2 ){ 
		for(i=0; i<Dim.num_regr; i++){
			for(j=i; j<Dim.num_regr; j++){
				tmp_info[id_r+j][id_r+i] += -(-Q3_01 + Q4_01)*u_ik_beta[j]*(-Q3_01 + Q4_01)*u_ik_beta[i]/(Q*Q) + (-Q3_02 + Q4_02)*u_ik_beta[j]*u_ik_beta[i]/Q;
			}
		}

		if( t <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_01 - Q2_01)*u_ik_beta[j]*(-Q3_01 + Q4_01)*u_ik_beta[i]/(Q*Q);
				}
			}
		}

		if( s <= Dim.num_categ - 1 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(Q1_10 - Q3_10)*u_ij_beta[j]*(-Q3_01 + Q4_01)*u_ik_beta[i]/(Q*Q) - Q3_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}

		if( s >= 2 ){
			for(i=0; i<Dim.num_regr; i++){
				for(j=i; j<Dim.num_regr; j++){
					tmp_info[id_r+j][id_r+i] += -(-Q2_10 + Q4_10)*u_ij_beta[j]*(-Q3_01 + Q4_01)*u_ik_beta[i]/(Q*Q) + Q4_11*u_ik_beta[j]*u_ij_beta[i]/Q;
				}
			}
		}
	}


	for(i=0; i<Dim.num_categ-2; i++){
		for(j=i+1; j<Dim.num_categ-2; j++){
			tmp_info[j][i] *= 0.5;
		}
	}

	for(i=0; i<Dim.num_param; i++){
		  for(j=i+1; j<Dim.num_param; j++){
				 tmp_info[i][j] = tmp_info[j][i];
		  }
	}

	for(i=0; i<Dim.num_param; i++){
		  for(j=0; j<Dim.num_param; j++){
				info[i+j*Dim.num_param] = tmp_info[i][j];
		  }	
	}


	}
	
	
}
/*#endif*/


/*#if CV*/
//Cristiano Varin's implementation
void CV_score_info_log_prob_pair(const int position,
			      const int distance_neigh,
			      double *score,
			      double *info)
{
  int t=position;
  int l=distance_neigh;
  int i, j;
  double this_phi=Param.phi[index_dynamic];
  double prob_this_pair=0.0;
  double Z1, Z2, Z11, Z22, A, B;
  
  double phi_pow_l=R_pow_di(this_phi, l);
  double phi_pow_l_minus_one=R_pow_di(this_phi, l-1);
  
  double var=Param.sigma2+1;
  double sd=sqrt(var);
  double cov=Param.sigma2+phi_pow_l;
  double corr=cov/var;

  /* derivative of var wrt phi */
  double deriv_phi_var=0;
  /* derivative of cov wrt phi  */
  double deriv_phi_cov=(l*phi_pow_l_minus_one);
  /* derivative of corr wrt phi and sigma2 */
  double deriv_phi_corr=(deriv_phi_cov*var-cov*deriv_phi_var)/R_pow_di(var,2); 
  double deriv_sigma2_corr=(var-cov)/R_pow_di(var,2); 

  for(i=0; i<Dim.num_param; i++)
    score[i]=0.0;


	double tmp_info[Dim.num_param][Dim.num_param];
  for(i=0; i<Dim.num_param; i++){
	  for(j=0; j<Dim.num_param; j++){
			tmp_info[i][j] = 0.0;
	  }
	}
	//
  prob_pair(t, l, &prob_this_pair);

  /* split the Gaussian rectangular probability into
     four cumulative Gaussian distribution functions */
  
  /* the next lines are written in a comfortable but not efficient order! */
  
  /* first term */
  if (data[t]<Dim.num_categ)
    Z1=(Param.tau[data[t]-1]-means[t])/sd;
  else
    Z1=BIG;
  if(data[t-l]<Dim.num_categ)
    Z2=(Param.tau[data[t-l]-1]-means[t-l])/sd;
  else
    Z2=BIG;

  A=dnorm(Z1, 0.0, 1.0, 0)*pnorm(((Z2-corr*Z1)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
  B=dnorm(Z2, 0.0, 1.0, 0)*pnorm(((Z1-corr*Z2)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
	//NL
	//the cathegories in data go from 1 (not 0!) to h
	//
  /* cutpoints */  
  if(data[t]>=2 && data[t]<=(Dim.num_categ-1)){
    score[data[t]-2]+=A/sd;
  }
  if(data[t-l]>=2 && data[t-l]<=(Dim.num_categ-1)){
    score[data[t-l]-2]+=B/sd;
  }
  /* regressors */
  for(i=0; i<Dim.num_regr; i++)
    score[i+Dim.num_categ-2]=-(A*design_matrix[t][i]/sd+
			       B*design_matrix[t-l][i]/sd);
  /* autocorrelation */
  score[Dim.num_categ+Dim.num_regr-2+index_dynamic]=
    A*(-Z1/(2*var)*deriv_phi_var)+B*(-Z2/(2*var)*deriv_phi_var)
    +dnorm2(Z1, Z2, corr)*deriv_phi_corr;
  /* random effect */
  if(Dim.is_random_effect)
    score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr]=
      A*(-Z1/(2*var))+B*(-Z2/(2*var))+dnorm2(Z1, Z2, corr)*
      deriv_sigma2_corr;
				    
  /* second term */
  if(data[t]>=2){
    Z11=(Param.tau[data[t]-2]-means[t])/sd;
      
    A=dnorm(Z11, 0.0, 1.0, 0)*pnorm(((Z2-corr*Z11)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
    B=dnorm(Z2, 0.0, 1.0, 0)*pnorm(((Z11-corr*Z2)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
    /* cutpoints */  
    if(data[t]>=3 && data[t]<=Dim.num_categ){
      score[data[t]-3]-=A/sd;
		}
    if(data[t-l]>=2 && data[t-l]<=(Dim.num_categ-1)){
      score[data[t-l]-2]-=B/sd;
    }
  /* regressors */      
    for(i=0; i<Dim.num_regr; i++)
      score[i+Dim.num_categ-2]-=-(A*design_matrix[t][i]/sd+
				     B*design_matrix[t-l][i]/sd);
  /* autocorrelation */
    score[Dim.num_categ+Dim.num_regr-2+index_dynamic]-=
      A*(-Z11/(2*var)*deriv_phi_var)+B*(-Z2/(2*var)*deriv_phi_var)
      +dnorm2(Z11, Z2, corr)*deriv_phi_corr;
  /* random effect */
    if(Dim.is_random_effect)
      score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr]-=
	A*(-Z11/(2*var))+B*(-Z2/(2*var))+dnorm2(Z11, Z2, corr)*
	deriv_sigma2_corr;
  }
      
  /* third term */
  if(data[t-l]>=2){
    Z22=(Param.tau[data[t-l]-2]-means[t-l])/sd;
    
    A=dnorm(Z1, 0.0, 1.0, 0)*pnorm(((Z22-corr*Z1)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
    B=dnorm(Z22, 0.0, 1.0, 0)*pnorm(((Z1-corr*Z22)/sqrt(1.0-corr*corr)), 0.0, 1.0, 1, 0);
	  /* cutpoints */      
    if(data[t]>=2 && data[t]<=(Dim.num_categ-1)){
      score[data[t]-2]-=A/sd;
    }
    if(data[t-l]>=3 && data[t-l]<=Dim.num_categ){
      score[data[t-l]-3]-=B/sd;
    }
  	/* regressors */                  
    for(i=0; i<Dim.num_regr; i++)
      score[i+Dim.num_categ-2]-=-(A*design_matrix[t][i]/sd+
				     B*design_matrix[t-l][i]/sd);
		/* autocorrelation */
    score[Dim.num_categ+Dim.num_regr-2+index_dynamic]-=
      A*(-Z1/(2*var)*deriv_phi_var)+B*(-Z22/(2*var)*deriv_phi_var)
      +dnorm2(Z1, Z22, corr)*deriv_phi_corr;
  	/* random effect */
    if(Dim.is_random_effect)
      score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr]-=
	A*(-Z1/(2*var))+B*(-Z22/(2*var))+dnorm2(Z1, Z22, corr)*
	deriv_sigma2_corr;
  }

  /* fourth term */
  if(data[t]>=2 && data[t-l]>=2){

    A=dnorm(Z11, 0.0, 1.0, 0)*pnorm(((Z22-corr*Z11)/sqrt(1-corr*corr)), 0.0, 1.0, 1, 0);
    B=dnorm(Z22, 0.0, 1.0, 0)*pnorm(((Z11-corr*Z22)/sqrt(1-corr*corr)), 0.0, 1.0, 1, 0);
	  /* cutpoints */      
    if(data[t]>=3 && data[t]<=Dim.num_categ){
      score[data[t]-3]+=A/sd;
    }
    if(data[t-l]>=3 && data[t-l]<=Dim.num_categ){
      score[data[t-l]-3]+=B/sd;
    }
  	/* regressors */                     
    for(i=0; i<Dim.num_regr; i++)
      score[i+Dim.num_categ-2]+=-(A*design_matrix[t][i]/sd+
				     B*design_matrix[t-l][i]/sd);
		/* autocorrelation */
    score[Dim.num_categ+Dim.num_regr-2+index_dynamic]+=
      A*(-Z11/(2*var)*deriv_phi_var)+B*(-Z22/(2*var)*deriv_phi_var)
      +dnorm2(Z11, Z22, corr)*deriv_phi_corr;
  /* random effect */
    if(Dim.is_random_effect)
      score[(Dim.num_categ-2)+Dim.num_regr+Dim.num_auto_corr]+=
	A*(-Z11/(2*var))+B*(-Z22/(2*var))+dnorm2(Z11, Z22, corr)*
	deriv_sigma2_corr;
  }
  
  for(i=0; i<Dim.num_param; i++)
    score[i]/=prob_this_pair;

   for(i=0; i<Dim.num_param; i++)
      for(j=0; j<Dim.num_param; j++)
	info[i+j*Dim.num_param]=score[i]*score[j];

}
/*#endif*/

void score_info_log_prob_pair(const int position,
			      const int distance_neigh,
			      double *score,
			      double *info,
			      int approx_hessian){
			      
	if(approx_hessian)
		CV_score_info_log_prob_pair(position, distance_neigh, score, info);
	else
		NL_score_info_log_prob_pair(position, distance_neigh, score, info);     
}


#endif

