### Name: maop
### Title: Mixed Autoregressive Ordered Probit Models
### Aliases: maop print.maop summary.maop
### Keywords: Longitudinal data, ordinal categorical response, pairwise
###   likelihood, random effects.

### ** Examples

# Fits a MAOP model to migraine data (Varin and Czado, 2009)
library(maop)
data(migraine)
attach(migraine)
# covariates
university <-rep(0,length(HA))
university[ED==5]<-1
university[ED==6]<-1
university[ED==7]<-1
university <- as.factor(university==1)
analgesics <- as.factor(MEDANALG=="Y")
# weather covariates
wc <- factor(cut(MINWC, breaks=c(-50, -10, 0, 10, 30)))
high.pressure <- PMN > 10130
high.pressure.yesterday <- PMN1P > 10130
change <- rep("persistence", length(high.pressure))
change[(!high.pressure & high.pressure.yesterday)] <- "worse"
change[(high.pressure & !high.pressure.yesterday)] <- "better"
change <- factor(change)
humidity <- cut(RHMN, breaks=c(0, 60, 80, 100))
Grad.tol.conv <- 1e-2




base <- HA~university+analgesics
all.models <- list(NULL)
all.models[[1]] <- base
all.models[[2]] <- update(base, .~. + change)
all.models[[3]] <- update(base, .~. + humidity)
all.models[[4]] <- update(base, .~. + wc)
all.models[[5]] <- update(base, .~. + change + humidity)
all.models[[6]] <- update(base, .~. + change + wc)
all.models[[7]] <- update(base, .~. + humidity + wc)
all.models[[8]] <- update(base, .~. + change + humidity + wc)



res.pw <- res.pw.rbr <- vector(len=length(all.models), mode="list")
mod.names <- unlist(strsplit(as.character(all.models),"~ "))[unlist(strsplit(as.character(all.models),"~ "))!="HA "]
names(res.pw) <- names(res.pw.rbr) <- mod.names

for(i in 1:length(all.models)){
	print(i)
	#estimation based on pairwise likelihood
		tmp <- maop(all.models[[i]], subjects=SUBJECT, periods=PERIOD, randomEffect=TRUE, dynamics=analgesics, distance=12, reparam=TRUE, grad.tol.conv=Grad.tol.conv)

		if(any(is.na(tmp$coeff)))
			tmp <- maop(all.models[[i]], subjects=SUBJECT, periods=PERIOD, randomEffect=TRUE, dynamics=analgesics, distance=12, reparam=TRUE, grad.tol.conv=Grad.tol.conv, all.methods=TRUE)
		
		res.pw[[i]] <- tmp

		save.image(file="migraine_estimated_models.rda")

	#estimation based on pairwise likelihood with empirical bias reduction
		tmp <- maop(all.models[[i]], subjects=SUBJECT, periods=PERIOD, randomEffect=TRUE, dynamics=analgesics, distance=12, reparam=TRUE, grad.tol.conv=Grad.tol.conv, RBR=TRUE, startValues=res.pw[[i]]$coeff)

		if(any(is.na(tmp$coeff)))
			tmp <- maop(all.models[[i]], subjects=SUBJECT, periods=PERIOD, randomEffect=TRUE, dynamics=analgesics, distance=12, reparam=TRUE, grad.tol.conv=Grad.tol.conv, RBR=TRUE, all.methods=TRUE, startValues=res.pw[[i]]$coeff)
		
		res.pw.rbr[[i]] <- tmp

		save.image(file="migraine_estimated_models.rda")
}




res.clic <- clic.weights <- delta.k <- matrix(NA, nr=length(res.pw), nc=2)

rownames(res.clic) <- rownames(clic.weights) <- rownames(delta.k) <- mod.names
colnames(res.clic) <- colnames(clic.weights) <- colnames(delta.k) <- c("pw","pen.pw")

res.clic[,1] <- unlist(lapply(res.pw, function(u) u$clic))
res.clic[,2] <- unlist(lapply(res.pw.rbr, function(u) u$clic))


delta.k[,1] <- (res.clic[,1]-min(res.clic[,1]))/2
delta.k[,2] <- (res.clic[,2]-min(res.clic[,2]))/2

clic.weights[,1] <- exp(-delta.k[,1])/sum(exp(-delta.k[,1]))
clic.weights[,2] <- exp(-delta.k[,2])/sum(exp(-delta.k[,2]))


#table 6
table6 <- cbind(res.clic, clic.weights)
table6 <- round(table6[,c(1,3,2,4)],2)

#table 7
##pairwise log lik
coeff <- res.pw[[1]]$coeff
stdErr <- res.pw[[1]]$stdErr
varCov <-  res.pw[[1]]$varMatrix
	tmp.coeff <- c(coeff[1:7], NA, NA, coeff[8], coeff[9], coeff[9]-coeff[8], coeff[10])
	tmp.stdErr <- c(stdErr[1:7], NA, NA, stdErr[8], stdErr[9], sqrt(varCov[8,8]+varCov[9,9]-2*varCov[8,9]), stdErr[10])
table7 <- cbind(tmp.coeff, tmp.stdErr)

coeff <- res.pw[[2]]$coeff
stdErr <- res.pw[[2]]$stdErr
varCov <-  res.pw[[2]]$varMatrix
	tmp.coeff <- c(coeff[-12], coeff[11]-coeff[10], coeff[12])
	tmp.stdErr <- c(stdErr[-12],sqrt(varCov[10,10]+varCov[11,11]-2*varCov[10,11]), stdErr[12])
table7 <- cbind(table7, tmp.coeff, tmp.stdErr)


##penalized pairwise log lik
coeff <- res.pw.rbr[[1]]$coeff
stdErr <- res.pw.rbr[[1]]$stdErr
varCov <-  res.pw.rbr[[1]]$varMatrix
	tmp.coeff <- c(coeff[1:7], NA, NA, coeff[8], coeff[9], coeff[9]-coeff[8], coeff[10])
	tmp.stdErr <- c(stdErr[1:7], NA, NA, stdErr[8], stdErr[9], sqrt(varCov[8,8]+varCov[9,9]-2*varCov[8,9]), stdErr[10])
table7 <- cbind(table7, tmp.coeff, tmp.stdErr)

coeff <- res.pw.rbr[[2]]$coeff
stdErr <- res.pw.rbr[[2]]$stdErr
varCov <-  res.pw.rbr[[2]]$varMatrix
	tmp.coeff <- c(coeff[-12], coeff[11]-coeff[10], coeff[12])
	tmp.stdErr <- c(stdErr[-12],sqrt(varCov[10,10]+varCov[11,11]-2*varCov[10,11]), stdErr[12])
table7 <- round(cbind(table7, tmp.coeff, tmp.stdErr),3)


save.image(file="migraine_estimated_models.rda")
