## Copyright (C) 2020 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Date: 06 January 2020
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

## overdispersion parameter is phi
psi2phi <- function(psi) (1 + psi)/psi
## psi in (0, infty) (see latent model)
phi2psi <- function(phi) 1/(phi - 1)

## This is in beta, log(phi) parameterization
loglik <- function(pars, y, x) {
    p <- ncol(x)
    beta <- pars[-(p + 1)]
    eta <- drop(x %*% beta)
    mu <- exp(eta)
    phi <- exp(pars[(p + 1)])
    psi <- phi2psi(phi)
    theta <- mu * psi    
    out <- sum(lgamma(theta + y) - lgamma(theta) + y * log(mu) + theta * log(theta) - (theta + y) * log(theta + mu))
    if (is.na(out) | is.infinite(out)) {
        -Inf
    }
    else {
        out
    }
}

## This is in beta, phi parameterization
simulate_nb <- function(pars, x) {
    n <- nrow(x)
    p <- ncol(x)
    beta <- pars[-(p + 1)]
    eta <- drop(x %*% beta)
    mu <- exp(eta)
    phi <- pars[(p + 1)]
    psi <- phi2psi(phi)
    theta <- mu * psi    
    rnegbin(n, mu, theta)
}

## This returns results in beta, phi parameterization
get_ml <- function(start, y, x, use_optim = FALSE, epsilon = 1e-02) {
    p <- ncol(x)
    if (use_optim) {
        out <- try(optim(start, fn = loglik, x = x, y = y, method = "BFGS",
                         control = list(fnscale = -1, maxit = 1000)), silent = TRUE)
    }
    else {
        out <- try(nlminb(start, objective = function(pars) -loglik(pars, x = x, y = y)),
                   silent = TRUE)
    }
    if (inherits(out, "try-error")) {
        coefs <- rep(NA_real_, p + 1)
        obj <- NA_real_
        gradient <- rep(NA_real_, p + 1)
        converged <- FALSE
    }
    else {
        obj <- if (use_optim) -out$value else -out$objective
        coefs <- out$par
        gradient <- try(grad(function(pars) loglik(pars, y = y, x = x), x = coefs), silent = TRUE)
        if (inherits(gradient, "try-error")) {
            gradient <- rep(NA_real_, length(coefs))
            converged <- FALSE
        }
        else {
            coefs[p + 1] <- exp(coefs[p + 1])
            gradient[p + 1] <- gradient[p + 1] / coefs[p + 1]
            converged <- isTRUE(all(abs(gradient) < epsilon))
        }
    }
    list(coefficients = coefs, loglik = obj, gradient = gradient, converged = converged)
}

## variance function is phi * mu
fit1_br <- function(y, x, slowit = 0.3, maxit = 10000, epsilon = 1e-04, alt_start = NULL) {
    dat <- data.frame(y = y, x = x)
    colnames(dat) <- c("y", "x1", "x2", "x3")
    formula0 <- y ~ -1 + x1 + x2 + x3
    c_names <- c("x1", "x2", "x3", "dispersion")
    n_par <- length(c_names)
    methods <- c("ML", "QL", "QLn", "BR", "BRn", "BRnp")
    n_methods <- length(methods)
    
    ql <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                  method = "brquasi",
                  slowit = slowit, maxit = maxit, epsilon = epsilon,
                  only_beta = FALSE,
                  type = "QL",
                  disp_factor = "n-p"), silent = TRUE)
    if (inherits(ql, "try-error")) {
        ql <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                      method = "brquasi",
                      slowit = slowit, maxit = maxit, epsilon = epsilon,
                      start = alt_start,
                      only_beta = FALSE,
                      type = "QL",
                      disp_factor = "n-p"), silent = TRUE)
        ## If we still fail
        if (inherits(ql, "try-error")) {
            out <- data.frame(parameter = rep(c_names, n_methods),
                              method = rep(methods, each = n_par),
                              converged = rep(rep(FALSE, n_methods), each = n_par),
                              estimate = rep(NA, n_methods * n_par),
                              score = rep(NA, n_methods * n_par))
            return(out)
        }
    }
    cql <- coef(ql, model = "full")
    
    qln <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                   method = "brquasi",
                   slowit = slowit, maxit = maxit, epsilon = epsilon,
                   start = cql,
                   type = "QL", 
                   only_beta = FALSE,
                   disp_factor = "n"), silent = TRUE)
    
    br <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                  method = "brquasi",
                  slowit = slowit, maxit = maxit, epsilon = epsilon,
                  start = cql,
                  type = "AS_empirical_mean", 
                  only_beta = TRUE,
                  disp_factor = "n-p"), silent = TRUE)
    brn <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                   method = "brquasi",
                   slowit = slowit, maxit = maxit, epsilon = epsilon,
                   start = cql,
                   type = "AS_empirical_mean", 
                   only_beta = FALSE,
                   disp_factor = "n"), silent = TRUE)

    brnp <- try(glm(formula0, family = quasi(link = "log", variance = "mu"), data = dat,
                    method = "brquasi",
                    slowit = slowit, maxit = maxit, epsilon = epsilon,
                    start = cql,
                    type = "AS_empirical_mean", 
                    only_beta = FALSE,
                    disp_factor = "n-p"), silent = TRUE)
    
    ml <- get_ml(start = cql, y = y, x = x, use_optim = FALSE)
    if (!ml$converged) {
        ml <- get_ml(start = cql, y = y, x = x, use_optim = TRUE)
    }
    if (!ml$converged) {
        ml <- try(get_ml(start = alt_start, y = y, x = x, use_optim = FALSE))
    }
    if (!ml$converged) {
        ml <- get_ml(start = alt_start, y = y, x = x, use_optim = TRUE)
    }

    failed_br <- inherits(br, "try-error")
    failed_qln <- inherits(qln, "try-error")
    failed_brn <- inherits(brn, "try-error")
    failed_brnp <- inherits(brnp, "try-error")
    
    cqln <- if (failed_qln) rep(NA, n_par) else coef(qln, model = "full")
    cbr <- if (failed_br) rep(NA, n_par) else coef(br, model = "full")
    cbrn <- if (failed_brn) rep(NA, n_par) else coef(brn, model = "full")
    cbrnp <- if (failed_brnp) rep(NA, n_par) else coef(brnp, model = "full")
    cml <- ml$coefficients
    
    converged <- c(ml$converged,
                   ql$converged,
                   if (failed_qln) FALSE else qln$converged,
                   if (failed_br) FALSE else br$converged,
                   if (failed_brn) FALSE else brn$converged,
                   if (failed_brnp) FALSE else brnp$converged)

    
    grads <- c(ml$gradient,
               ql$grad,
               if (failed_qln) rep(NA, n_par) else qln$grad,
               if (failed_br) rep(NA, n_par) else br$grad,
               if (failed_brn) rep(NA, n_par) else brn$grad,
               if (failed_brnp) rep(NA, n_par) else brnp$grad)
    
    data.frame(parameter = rep(c_names, n_methods),
               method = rep(methods, each = n_par),
               converged = rep(converged, each = n_par),
               estimate = c(cml, cql, cqln, cbr, cbrn, cbrnp),
               score = grads)
}


fit2_br <- function(y, x, slowit = 0.3, maxit = 10000, epsilon = 1e-04, alt_start = NULL) {
    dat <- data.frame(y = y, x = x)
    colnames(dat) <- c("y", "x1", "x2", "x3")
    formula0 <- y ~ -1 + x1 + x2 + x3
    c_names <- c("x1", "x2", "x3", "dispersion")
    n_par <- length(c_names)
    methods <- c("ML", "QL", "QLn", "BR", "BRn")
    n_methods <- length(methods)
    
    ql <- try(glm(formula0, family = quasi(link = "log", variance = "mu^2"), data = dat,
                  method = "brquasi",
                  slowit = slowit, maxit = maxit, epsilon = epsilon,
                  only_beta = FALSE,
                  type = "QL",
                  disp_factor = "n-p"), silent = TRUE)
    if (inherits(ql, "try-error")) {
        ql <- try(glm(formula0, family = quasi(link = "log", variance = "mu^2"), data = dat,
                      method = "brquasi",
                      slowit = slowit, maxit = maxit, epsilon = epsilon,
                      start = alt_start,
                      only_beta = FALSE,
                      type = "QL",
                      disp_factor = "n-p"), silent = TRUE)
        ## If we still fail
        if (inherits(ql, "try-error")) {
            out <- data.frame(parameter = rep(c_names, n_methods),
                              method = rep(methods, each = n_par),
                              converged = rep(rep(FALSE, n_methods), each = n_par),
                              estimate = rep(NA, n_methods * n_par),
                              score = rep(NA, n_methods * n_par))
            return(out)
        }
    }
    cql <- coef(ql, model = "full")
    
    qln <- try(glm(formula0, family = quasi(link = "log", variance = "mu^2"), data = dat,
                   method = "brquasi",
                   slowit = slowit, maxit = maxit, epsilon = epsilon,
                   start = cql,
                   type = "QL", 
                   only_beta = FALSE,
                   disp_factor = "n"), silent = TRUE)
    
    br <- try(glm(formula0, family = quasi(link = "log", variance = "mu^2"), data = dat,
                  method = "brquasi",
                  slowit = slowit, maxit = maxit, epsilon = epsilon,
                  start = cql,
                  type = "AS_empirical_mean", 
                  only_beta = TRUE,
                  disp_factor = "n-p"), silent = TRUE)

    brn <- try(glm(formula0, family = quasi(link = "log", variance = "mu^2"), data = dat,
                   method = "brquasi",
                   slowit = slowit, maxit = maxit, epsilon = epsilon,
                   start = cql,
                   type = "AS_empirical_mean", 
                   only_beta = FALSE,
                   disp_factor = "n"), silent = TRUE)

    ml <- get_ml(start = cql, y = y, x = x, use_optim = FALSE)
    if (!ml$converged) {
        ml <- get_ml(start = alt_start, y = y, x = x, use_optim = FALSE)
    }
    if (!ml$converged) {
        ml <- get_ml(start = cql, y = y, x = x, use_optim = TRUE)
    }
    if (!ml$converged) {
        ml <- get_ml(start = alt_start, y = y, x = x, use_optim = TRUE)
    }

    failed_br <- inherits(br, "try-error")
    failed_qln <- inherits(qln, "try-error")
    failed_brn <- inherits(brn, "try-error")
    
    cqln <- if (failed_qln) rep(NA, n_par) else coef(qln, model = "full")
    cbr <- if (failed_br) rep(NA, n_par) else coef(br, model = "full")
    cbrn <- if (failed_brn) rep(NA, n_par) else coef(brn, model = "full")
    cml <- ml$coefficients
    
    converged <- c(ml$converged,
                   ql$converged,
                   if (failed_qln) FALSE else qln$converged,
                   if (failed_br) FALSE else br$converged,
                   if (failed_brn) FALSE else brn$converged)

    grads <- c(ml$gradient,
               ql$grad,
               if (failed_qln) rep(NA, n_par) else qln$grad,
               if (failed_br) rep(NA, n_par) else br$grad,
               if (failed_brn) rep(NA, n_par) else brn$grad)
    
    data.frame(parameter = rep(c_names, n_methods),
               method = rep(methods, each = n_par),
               converged = rep(converged, each = n_par),
               estimate = c(cml, cql, cqln, cbr, cbrn),
               score = grads)
}

