## Copyright (C) 2020 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Date: 06 January 2020
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

library("parallel")
library("dplyr")
library("ggplot2")
library("data.table")

code_path <- "./"
path <- "./images/" 
output_path <- "./"

source(file.path(code_path, "negbin_functions.R"))

load(file.path(path, "negbin-summaries_mu.rda"))
load(file.path(path, "negbin-summaries_musquared.rda"))

## Remove dispersion results from musquared
summaries_musquared <- summaries_musquared %>% filter(parameter != "dispersion")

## Add indicator and re-order facets
summaries_mu$V <- "V(mu) == mu"
summaries_musquared$V <- "V(mu) == mu^2"
summaries <- rbind(summaries_mu, summaries_musquared)

summaries$method <- factor(summaries$method, levels = c("ML", "QL", "QLn", "BRn", "BR"), ordered = TRUE)
summaries$method <- recode(summaries$method, "BRn" = "RBM:paste(beta,phi)", "BR" = "RBM:beta")
summaries$parameter <- recode(summaries$parameter, "x1" = "beta[0]", "x2" = "beta[1]", "x3" = "beta[2]", "dispersion" = "phi")
## s_summaries <- summaries %>% filter(!((method == "ML") & (V == "V(mu) == mu^2")), method != "QLn", n < 2500)
s_summaries <- summaries %>% filter(method != "QLn", n < 2500)


quantile <- qnorm(1 - 0.01/2)


ylims <- c(-2.7, 2.7)
pdf(paste0(output_path, "/", "brquasi_scaled_bias_beta0.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[0]")) +
    geom_hline(aes(yintercept = 0), col = "grey") +
    geom_point(aes(x = log(n), y = scaled_bias),
               position = position_dodge(width = 0.5), size = 0.5) +
    geom_errorbar(aes(x = log(n), 
                      ymin = scaled_bias - quantile * scaled_bias_mc_error,
                      ymax = scaled_bias + quantile * scaled_bias_mc_error),
                  position = position_dodge(width = 0.5)) +
    facet_grid(V ~ method, scales = "free", labeller = label_parsed) +
    theme_minimal() +
    labs(x = expression(log(n[r])), y = expression(n[r] %*% Bias), title = expression(beta[0]), labeller = label_parsed) + 
    theme(legend.position = "none") +
    coord_cartesian(ylim = ylims) 
dev.off()

pdf(paste0(output_path, "/", "brquasi_scaled_bias_beta1.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[1]")) +
    geom_hline(aes(yintercept = 0), col = "grey") +
    geom_point(aes(x = log(n), y = scaled_bias),
               position = position_dodge(width = 0.5), size = 0.5) +
    geom_errorbar(aes(x = log(n), 
                      ymin = scaled_bias - quantile * scaled_bias_mc_error,
                      ymax = scaled_bias + quantile * scaled_bias_mc_error),
                  position = position_dodge(width = 0.5)) +
    facet_grid(V ~ method, scales = "free", labeller = label_parsed) +
    theme_minimal() +
    labs(x = expression(log(n[r])), y = NULL, title = expression(beta[1]), labeller = label_parsed) + 
    theme(legend.position = "none") +
    coord_cartesian(ylim = ylims)
dev.off()

pdf(paste0(output_path, "/", "brquasi_scaled_bias_beta2.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[2]")) +
    geom_hline(aes(yintercept = 0), col = "grey") +
    geom_point(aes(x = log(n), y = scaled_bias),
               position = position_dodge(width = 0.5), size = 0.5) +
    geom_errorbar(aes(x = log(n), 
                      ymin = scaled_bias - quantile * scaled_bias_mc_error,
                      ymax = scaled_bias + quantile * scaled_bias_mc_error),
                  position = position_dodge(width = 0.5)) +
    facet_grid(V ~ method, scales = "free", labeller = label_parsed) +
    theme_minimal() +
    labs(x = expression(log(n[r])), y = NULL, title = expression(beta[2]), labeller = label_parsed) + 
    theme(legend.position = "none") +
    coord_cartesian(ylim = ylims)    
dev.off()

pdf(paste0(output_path, "/", "brquasi_scaled_bias_phi.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "phi")) +
    geom_hline(aes(yintercept = 0), col = "grey") +
    geom_point(aes(x = log(n), y = scaled_bias),
               position = position_dodge(width = 0.5), size = 0.5) +
    geom_errorbar(aes(x = log(n), 
                      ymin = scaled_bias - quantile * scaled_bias_mc_error,
                      ymax = scaled_bias + quantile * scaled_bias_mc_error),
                  position = position_dodge(width = 0.5)) +
    facet_grid(V ~ method, scales = "free", labeller = label_parsed) +
    theme_minimal() +
    labs(x = expression(log(n[r])), y = expression(n[r] %*% Bias), title = expression(phi), labeller = label_parsed) + 
    theme(legend.position = "none") +
    coord_cartesian(ylim = c(-20, 5))
dev.off()


pdf(paste0(output_path, "/", "brquasi_perc_converged.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[0]")) +
    geom_col(aes(x = log(n), y = perc_converged)) +
    facet_grid(V ~ method, scales = "free", labeller = label_parsed) +
    theme_minimal() +
    theme(legend.position = "none") +
    labs(x = expression(log(n[r])), y = expression("% converged"), title = "", labeller = label_parsed) + 
    coord_cartesian(ylim = c(0, 100))
dev.off()


pdf(paste0(output_path, "/", "brquasi_variance_beta0.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[0]")) +
    geom_col(aes(x = log(n), y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    facet_grid(V ~ ., labeller = label_parsed) +
    theme_minimal() +
    theme(legend.position = "none") +
    labs(x = expression(log(n[r])), y = expression("variance"), title = expression(beta[0]), labeller = label_parsed)  +
    scale_colour_grey() +
    scale_fill_grey()
dev.off()


pdf(paste0(output_path, "/", "brquasi_variance_beta1.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[1]")) +
    geom_col(aes(x = log(n), y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    facet_grid(V ~ ., labeller = label_parsed) +
    theme_minimal() +
    theme(legend.position = "none") +
    labs(x = expression(log(n[r])), y = expression("variance"), title = expression(beta[1]), labeller = label_parsed)  +
    scale_colour_grey() +
    scale_fill_grey()
dev.off()


pdf(paste0(output_path, "/", "brquasi_variance_beta2.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "beta[2]")) +
    geom_col(aes(x = log(n), y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    facet_grid(V ~ ., labeller = label_parsed) +
    theme_minimal() +
    theme(legend.position = "none") +
    labs(x = expression(log(n[r])), y = expression("variance"), title = expression(beta[2]), labeller = label_parsed)  +
    scale_colour_grey() +
    scale_fill_grey()
dev.off()


pdf(paste0(output_path, "/", "brquasi_variance_phi.pdf"), width = 3, height = 3)
ggplot(s_summaries %>% filter(parameter == "phi")) +
    geom_col(aes(x = log(n), y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    facet_grid(V ~ ., labeller = label_parsed) +
    theme_minimal() +
    theme(legend.position = "none") +
    labs(x = expression(log(n[r])), y = expression("variance"), title = expression(phi), labeller = label_parsed)  +
    scale_colour_grey() +
    scale_fill_grey()
dev.off()
