## Copyright (C) 2020 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Date: 06 January 2020
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

library("parallel")
library("dplyr")
library("ggplot2")

## Path to read images from
path <- "./images/"

files <- dir(path, full.names = TRUE)
files <- files[grep("ratio-setting-copula", files)]

all_results <- NULL
for (f in files) {
    load(f)
    cres <- results
    cres$n <- attr(cres, "n")
    cres$N <- attr(cres, "N")
    all_results <- rbind(all_results, cres)
}

## Estimate threshold
M <- Inf

summaries <- NULL
## Summaries conditional on the finiteness of the ML estimate
for (j in seq_along(M)) {
    summaries[[j]] <- all_results %>%
        group_by(n, parameter, method) %>%
        ## exclude large estimates and only use estimates that are less than M in abolute value
        mutate(remove = estimate > M) %>%
        summarize(samples_perc = 100 * (1 - mean(remove)),
                  max_estimate = max(estimate),
                  bias = mean((estimate - true_value)),
                  mad = mean(abs(estimate - true_value)),
                  pu = mean(estimate < true_value),
                  variance = var(estimate),
                  est_variance = mean(evar),
                  mse = bias^2 + variance,
                  typeI_e = mean(pvalue_e < 0.05),
                  typeI_s = mean(pvalue_s < 0.05), 
                  N = n(),
                  mc_error = sd(estimate)/sqrt(N),
                  scaled_bias = bias * unique(n),
                  scaled_bias_mc_error = mc_error * unique(n),
                  M = M[j])
}

summaries <- summaries[[1]]




pdf(paste0(path, "/ratio-setting-3.pdf"))

summaries$method <- factor(summaries$method, levels = c("ML", "BR", "Jackknife"), ordered = TRUE)
ggplot(summaries) +
    geom_col(aes(x = parameter, y = scaled_bias, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    theme_minimal() +
    facet_grid( ~ n)

ggplot(summaries) +
    geom_col(aes(x = parameter, y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    theme_minimal() +
    ## coord_flip() +
    facet_grid( ~ n)

ggplot(summaries) +
    geom_line(aes(x = log(n, 2), y = log(abs(bias), 2), color = method)) +
    theme_minimal() +
    facet_grid(~ parameter)

ggplot(summaries) +
    geom_line(aes(x = log(n, 2), y = samples_perc, color = method)) +
    coord_cartesian(ylim = c(0, 100)) +
    theme_minimal() +
    facet_grid(~ parameter)


dev.off()


## Output for latex

what <- c("bias", "mse", "mad", "pu")

tab <- NULL
summaries_wide <- NULL
for (w in what) {
    summs <- as.data.frame(summaries[, c("n", "method", w)])
    summs <- reshape(summs, v.names = w, direction = "wide", idvar = "method", timevar ="n")
    summaries_wide[[w]] <- summs
    summs <- data.frame(summary = w, summs, stringsAsFactors = FALSE)
    nams <- names(summs)
    names(summs) <- gsub(paste0(w, "."), "n=", nams)
    summs <- summs[c(3, 2, 1), ]
    rownames(summs) <- NULL
    tab <- rbind(tab, summs)
}
names(tab)[which(names(tab) == "method")] <- "estimator"
names(tab)[which(names(tab) == "summary")] <- ""

memisc:::toLatex.data.frame(tab, useDcolumn = FALSE, digits = 2)

## Slopes

df <- reshape(summaries_wide$bias, direction = "long")
m_ML <- lm(I(log(abs(bias))) ~ log(n) + offset(-1 * log(n)), data = df, subset = method == "ML" & n > 20)
m_BR <- lm(I(log(abs(bias))) ~ log(n) + offset(-2 * log(n)), data = df, subset = method == "BR" & n > 20)
m_Ja <- lm(I(log(abs(bias))) ~ log(n) + offset(-2 * log(n)), data = df, subset = method == "Jackknife" & n > 20)

## Variances
what <- c("variance", "est_variance")

tab <- NULL
summaries_wide <- NULL
for (w in what) {
    summs <- as.data.frame(summaries[, c("n", "method", w)] %>% filter(method != "Jackknife", n > 2))
    summs <- reshape(summs, v.names = w, direction = "wide", idvar = "method", timevar ="n")
    summaries_wide[[w]] <- summs
    summs <- data.frame(summary = w, summs, stringsAsFactors = FALSE)
    nams <- names(summs)
    names(summs) <- gsub(paste0(w, "."), "n=", nams)
    summs <- summs[c(2, 1), ]
    rownames(summs) <- NULL
    tab <- rbind(tab, summs)
}
names(tab)[which(names(tab) == "method")] <- "estimator"
names(tab)[which(names(tab) == "summary")] <- ""

memisc:::toLatex.data.frame(tab, useDcolumn = FALSE, digits = 2)
